import numpy as np
from numpy.typing import NDArray
from typing import List, Tuple, Annotated, Optional, Union
from pydantic import validate_call, ConfigDict, BeforeValidator, Field
from cellprofiler_library.functions.measurement import measure_correlation_and_slope_from_objects, measure_manders_coefficient_from_objects, measure_rwc_coefficient_from_objects, measure_overlap_coefficient_from_objects, measure_costes_coefficient_from_objects, get_thresholded_images_and_counts, measure_correlation_and_slope, measure_manders_coefficient, measure_rwc_coefficient, measure_overlap_coefficient, measure_costes_coefficient
from cellprofiler_library.opts.measurecolocalization import TemplateMeasurementFormat, MeasurementType
from cellprofiler_library.types import ImageGrayscale, ImageGrayscaleMask, Pixel, ObjectLabel, ObjectSegmentation, ImageAny, ImageAnyMask
from cellprofiler_library.opts.measurecolocalization import CostesMethod
from cellprofiler_library.measurement_model import LibraryMeasurements
from cellprofiler_library.functions.image_processing import crop_image_similarly
from cellprofiler_library.functions.object_processing import size_similarly, object_crop_image_similarly

def crop_image_pair_similarly(
        im1_pixel_data:     Annotated[ImageGrayscale, Field(description="First image pixel data")],
        im2_pixel_data:     Annotated[ImageGrayscale, Field(description="Second image pixel data")],
        im1_mask:           Annotated[ImageGrayscaleMask, Field(description="First image mask")],
        im2_mask:           Annotated[ImageGrayscaleMask, Field(description="Second image mask")],
        im1_crop_mask:      Annotated[Optional[ImageGrayscaleMask], Field(description="Mask used to obtain the first image")]=None,
        im2_crop_mask:      Annotated[Optional[ImageGrayscaleMask], Field(description="Mask used to obtain the second image")]=None,
    ) -> Tuple[ImageGrayscale, ImageGrayscale, ImageGrayscaleMask]:
    
    im1_pixel_count = np.prod(im1_pixel_data.shape)
    im2_pixel_count = np.prod(im2_pixel_data.shape)

    if im1_pixel_count < im2_pixel_count:
        # im2_pixel_data = im1.crop_image_similarly(im2.pixel_data)
        im2_pixel_data = crop_image_similarly(im1_pixel_data, im2_pixel_data, im1_crop_mask)
        im2_mask = crop_image_similarly(im1_pixel_data, im2_mask, im1_crop_mask)
    elif im2_pixel_count < im1_pixel_count:
        im1_pixel_data = crop_image_similarly(im2_pixel_data, im1_pixel_data, im2_crop_mask)
        im1_mask = crop_image_similarly(im2_pixel_data, im1_mask, im2_crop_mask)
    mask = (
        im1_mask
        & im2_mask
        & (~np.isnan(im1_pixel_data))
        & (~np.isnan(im2_pixel_data))
    )
    return im1_pixel_data, im2_pixel_data, mask

def crop_image_pair_and_object_similarly(
        im1_pixel_data:     Annotated[ImageGrayscale, Field(description="First image pixel data")],
        im2_pixel_data:     Annotated[ImageGrayscale, Field(description="Second image pixel data")],
        obj_segmented:      Annotated[ObjectSegmentation, Field(description="Object segmentations")],
        im1_mask:           Annotated[ImageGrayscaleMask, Field(description="First image mask")],
        im2_mask:           Annotated[ImageGrayscaleMask, Field(description="Second image mask")],
        obj_parent_image:   Annotated[Optional[ImageAny], Field(description="Object parent image")],
        obj_parent_crop_mask: Annotated[Optional[ImageAnyMask], Field(description="Object parent image crop mask")]=None,
        im1_crop_mask:      Annotated[Optional[ImageGrayscaleMask], Field(description="Mask used to obtain the first image")]=None,
        im2_crop_mask:      Annotated[Optional[ImageGrayscaleMask], Field(description="Mask used to obtain the second image")]=None,
) -> Tuple[
    NDArray[Pixel],
    NDArray[Pixel],
    ObjectSegmentation,
    ImageGrayscaleMask,
    NDArray[Pixel],
    NDArray[Pixel]
]:
    try:
        im1_pixels = object_crop_image_similarly(obj_segmented, im1_pixel_data, obj_parent_image, obj_parent_crop_mask) # first check image shape and segmented shape, if new, crop image similarly wiht parent_image. To do that, verify parent image shape and the image shape, if neq, use parent image crop mask as well as the parent image shape 
        _im1_mask = object_crop_image_similarly(obj_segmented, im1_mask, obj_parent_image, obj_parent_crop_mask)
    except ValueError:
        im1_pixels, m1 = size_similarly(obj_segmented, im1_pixel_data)
        _im1_mask, m1 = size_similarly(obj_segmented, im1_mask)
        _im1_mask[~m1] = False
    try:
        im2_pixels = object_crop_image_similarly(obj_segmented, im2_pixel_data, obj_parent_image, obj_parent_crop_mask)
        _im2_mask = object_crop_image_similarly(obj_segmented, im2_mask, obj_parent_image, obj_parent_crop_mask)
    except ValueError:
        im2_pixels, m1 = size_similarly(obj_segmented, im2_pixel_data)
        _im2_mask, m1 = size_similarly(obj_segmented, im2_mask)
        _im2_mask[~m1] = False
    mask = (obj_segmented > 0) & _im1_mask & _im2_mask
    im1_pixels = im1_pixels[mask]
    im2_pixels = im2_pixels[mask]
    obj_segmented = obj_segmented[mask]
    #
    # Code below is used for the Costes' automated thresholding
    #
    # im1_mask = im1_mask
    im1_pixel_count = np.product(im1_pixel_data.shape)
    im2_pixel_data = im2_pixel_data
    # im2_mask = im2_mask
    im2_pixel_count = np.product(im2_pixel_data.shape)
    #
    # Crop the larger image similarly to the smaller one
    #
    if im1_pixel_count < im2_pixel_count:
        im2_pixel_data = crop_image_similarly(im1_pixel_data, im2_pixel_data, im1_crop_mask)
        im2_mask = crop_image_similarly(im1_pixel_data, im2_mask, im1_crop_mask)
    elif im2_pixel_count < im1_pixel_count:
        im1_pixel_data = crop_image_similarly(im2_pixel_data, im1_pixel_data, im2_crop_mask)
        im1_mask = crop_image_similarly(im2_pixel_data, im1_mask, im2_crop_mask)
    mask = (
        im1_mask
        & im2_mask
        & (~np.isnan(im1_pixel_data))
        & (~np.isnan(im2_pixel_data))
    )
    #
    # fi and si are used to obtain the costes threshold values for their respective images
    #
    im1_costes_pixels = None
    im2_costes_pixels = None
    if mask is not None and np.any(mask):
        im1_costes_pixels = im1_pixel_data[mask]
        im2_costes_pixels = im2_pixel_data[mask]

    return im1_pixels, im2_pixels, obj_segmented, mask, im1_costes_pixels, im2_costes_pixels

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def run_image_pair_images(
    im1_pixel_data:     Annotated[ImageGrayscale, Field(description="First image pixel data")],
    im2_pixel_data:     Annotated[ImageGrayscale, Field(description="Second image pixel data")],
    im1_name:           Annotated[str, Field(description="First image name")] = "Name of the first image. This will be used to name the measurement columns",
    im2_name:           Annotated[str, Field(description="Second image name")] = "Name of the second image. This will be used to name the measurement columns",
    mask:               Annotated[Optional[ImageGrayscaleMask], Field(description="Mask of the pixels to be used for the measurements")] = None, 
    im1_thr_percentage: Annotated[float, Field(description="Threshold value for the first image"), BeforeValidator(np.float64)]=100, 
    im2_thr_percentage: Annotated[float, Field(description="Threshold value for the second image"), BeforeValidator(np.float64)]=100, 
    measurement_types:  Annotated[List[MeasurementType], Field(description="List of measurement types to be calculated")] = [MeasurementType.CORRELATION, MeasurementType.MANDERS, MeasurementType.RWC, MeasurementType.OVERLAP, MeasurementType.COSTES],
    im1_scale:          Annotated[Optional[float], Field(description="")] = None,
    im2_scale:          Annotated[Optional[float], Field(description="")] = None,
    costes_method:      Annotated[Optional[CostesMethod], Field(description="")] = CostesMethod.FAST,
    ) -> Annotated[
        Tuple[
            LibraryMeasurements,
            List[Tuple[str, str, str, str, str]]
        ], Field(description="List of measurement results and a dictionary of measurements with precise values")]:
    """Calculate the correlation between the pixels of two images"""


    summary: List[Tuple[str, str, str, str, str]] = []
    measurements = LibraryMeasurements()
    corr =      np.float64(np.NaN)
    slope =     np.float64(np.NaN)
    C1 =        np.float64(np.NaN)
    C2 =        np.float64(np.NaN)
    M1 =        np.float64(np.NaN)
    M2 =        np.float64(np.NaN)
    RWC1 =      np.float64(np.NaN)
    RWC2 =      np.float64(np.NaN)
    overlap =   np.float64(np.NaN)
    K1 =        np.float64(np.NaN)
    K2 =        np.float64(np.NaN)
    if mask is not None and np.any(mask):
        im1_pixels = im1_pixel_data[mask]
        im2_pixels = im2_pixel_data[mask]

        if MeasurementType.CORRELATION in measurement_types:
            corr, slope = measure_correlation_and_slope(im1_pixels, im2_pixels)
            summary += [
                (im1_name, im2_name,"-","Correlation","%.3f" % corr),
                (im1_name, im2_name, "-", "Slope", "%.3f" % slope),
            ]

        if set(measurement_types).intersection({MeasurementType.MANDERS, MeasurementType.RWC, MeasurementType.OVERLAP, MeasurementType.COSTES}) != set():
            # Threshold as percentage of maximum intensity in each channel
            im1_thr_sum, im2_thr_sum, thr_mask_intersection = get_thresholded_images_and_counts(im1_pixels, im2_pixels, im1_thr_percentage, im2_thr_percentage)
            # take the first element as only a single value is expected for whole image threshold sum
            im1_thr_sum = im1_thr_sum[0]
            im2_thr_sum = im2_thr_sum[0]

            if MeasurementType.MANDERS in measurement_types:
                M1, M2 = measure_manders_coefficient(im1_pixels, im2_pixels, im1_thr_sum, im2_thr_sum, thr_mask_intersection)
                summary += [
                    (im1_name, im2_name, "-", "Manders Coefficient", "%.3f" % M1),
                    (im2_name, im1_name, "-", "Manders Coefficient", "%.3f" % M2),
                ]


            if MeasurementType.RWC in measurement_types:
                RWC1, RWC2 = measure_rwc_coefficient(im1_pixels, im2_pixels, im1_thr_sum, im2_thr_sum, thr_mask_intersection)
                summary += [
                    (im1_name, im2_name, "-", "RWC Coefficient", "%.3f" % RWC1),
                    (im2_name, im1_name, "-", "RWC Coefficient", "%.3f" % RWC2),
                ]

            if MeasurementType.OVERLAP in measurement_types:
                overlap, K1, K2 = measure_overlap_coefficient(im1_pixels, im2_pixels, thr_mask_intersection)
                summary += [
                    (im1_name, im2_name, "-", "Overlap Coefficient", "%.3f" % overlap),
                ]

            if MeasurementType.COSTES in measurement_types:
                assert costes_method is not None, "costes_method must be one of {}".format([e.value for e in CostesMethod])
                assert costes_method in [e.value for e in CostesMethod], "costes_method must be one of {}".format([e.value for e in CostesMethod])
                
                C1, C2 = measure_costes_coefficient(im1_pixels, im2_pixels, im1_scale, im2_scale, costes_method=costes_method)
                
                summary += [
                    (im1_name, im2_name, "-", "Manders Coefficient (Costes)", "%.3f" % C1),
                    (im2_name, im1_name, "-", "Manders Coefficient (Costes)", "%.3f" % C2),
                ]

    #
    # Add the measurements
    #
    corr_measurement = TemplateMeasurementFormat.CORRELATION_FORMAT % (im1_name, im2_name)
    slope_measurement = TemplateMeasurementFormat.SLOPE_FORMAT % (im1_name, im2_name)
    overlap_measurement = TemplateMeasurementFormat.OVERLAP_FORMAT % (im1_name, im2_name)

    k_measurement_1 = TemplateMeasurementFormat.K_FORMAT % (im1_name, im2_name)
    k_measurement_2 = TemplateMeasurementFormat.K_FORMAT % (im2_name, im1_name)
    
    manders_measurement_1 = TemplateMeasurementFormat.MANDERS_FORMAT % (im1_name, im2_name)
    manders_measurement_2 = TemplateMeasurementFormat.MANDERS_FORMAT % (im2_name, im1_name)

    rwc_measurement_1 = TemplateMeasurementFormat.RWC_FORMAT % (im1_name, im2_name)
    rwc_measurement_2 = TemplateMeasurementFormat.RWC_FORMAT % (im2_name, im1_name)

    costes_measurement_1 = TemplateMeasurementFormat.COSTES_FORMAT % (im1_name, im2_name)
    costes_measurement_2 = TemplateMeasurementFormat.COSTES_FORMAT % (im2_name, im1_name)
    
    if MeasurementType.CORRELATION in measurement_types:
        measurements.add_image_measurement(corr_measurement, corr)
        measurements.add_image_measurement(slope_measurement, slope)
    if MeasurementType.OVERLAP in measurement_types:        
        measurements.add_image_measurement(overlap_measurement, overlap)
        measurements.add_image_measurement(k_measurement_1, K1)
        measurements.add_image_measurement(k_measurement_2, K2)

    if MeasurementType.MANDERS in measurement_types:
        measurements.add_image_measurement(manders_measurement_1, M1)
        measurements.add_image_measurement(manders_measurement_2, M2)
    if MeasurementType.RWC in measurement_types:
        measurements.add_image_measurement(rwc_measurement_1, RWC1)
        measurements.add_image_measurement(rwc_measurement_2, RWC2)
    if MeasurementType.COSTES in measurement_types:
        measurements.add_image_measurement(costes_measurement_1, C1)
        measurements.add_image_measurement(costes_measurement_2, C2)

    return measurements, summary


def __get_object_result_array(col_order_list: Tuple[str, str, str], measurement_name: str, measurement_array: NDArray[np.float64]) -> List[Tuple[str, str, str, str, str]]:
    summary: List[Tuple[str, str, str, str, str]] = []
    summary += [
        (*col_order_list, f"Mean {measurement_name}", "%.3f" % np.mean(measurement_array)),
        (*col_order_list, f"Median {measurement_name}", "%.3f" % np.median(measurement_array)),
        (*col_order_list, f"Min {measurement_name}", "%.3f" % np.min(measurement_array)),
        (*col_order_list, f"Max {measurement_name}", "%.3f" % np.max(measurement_array)),
    ]
    return summary

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def run_image_pair_objects(
    im1_pixels:         Annotated[NDArray[Pixel], Field(description="First image pixels")],
    im2_pixels:         Annotated[NDArray[Pixel], Field(description="Second image pixels")],
    labels:             Annotated[NDArray[ObjectLabel], Field(description="Labels")],
    object_count:       Annotated[int, Field(description="Object count")],
    im1_name:           Annotated[str, Field(description="First image name")] = "First image",
    im2_name:           Annotated[str, Field(description="Second image name")] = "Second image",
    object_name:        Annotated[str, Field(description="Object name")] = "Objects",
    mask:               Annotated[Optional[ImageGrayscaleMask], Field(description="Mask")] = None, 
    im1_thr_percentage: Annotated[float, Field(description="First image threshold value"), BeforeValidator(np.float64)]=100, 
    im2_thr_percentage: Annotated[float, Field(description="Second image threshold value"), BeforeValidator(np.float64)]=100, 
    im1_costes_pixels:  Annotated[Optional[NDArray[Pixel]], Field(description="First image pixel data for costes")]=None,
    im2_costes_pixels:  Annotated[Optional[NDArray[Pixel]], Field(description="Second image pixel data for costes")]=None,
    measurement_types:  Annotated[List[MeasurementType], Field(description="List of measurement types to be calculated")]=[MeasurementType.CORRELATION, MeasurementType.MANDERS, MeasurementType.RWC, MeasurementType.OVERLAP, MeasurementType.COSTES],
    im1_scale:          Annotated[Optional[Union[float, int]], Field(description="First image scale for costes thresholding")]=None,
    im2_scale:          Annotated[Optional[Union[float, int]], Field(description="Second image scale for costes thresholding")]=None,
    costes_method:      Annotated[Optional[CostesMethod], Field(description="Costes method for costes thresholding")]=CostesMethod.FAST
    ) -> Tuple[
        LibraryMeasurements,
        List[Tuple[str, str, str, str, str]]
        ]:
    if MeasurementType.COSTES in measurement_types:
        assert costes_method is not None, "Costes requires a costes method"
    """Calculate per-object correlations between intensities in two images"""
    summary = []

    n_objects = object_count
    # Handle case when both images for the correlation are completely masked out
    corr = np.zeros((0,))
    overlap = np.zeros((0,))
    K1 = np.zeros((0,))
    K2 = np.zeros((0,))
    M1 = np.zeros((0,))
    M2 = np.zeros((0,))
    RWC1 = np.zeros((0,))
    RWC2 = np.zeros((0,))
    C1 = np.zeros((0,))
    C2 = np.zeros((0,))

    if n_objects == 0:
        corr = np.zeros((0,))
        overlap = np.zeros((0,))
        K1 = np.zeros((0,))
        K2 = np.zeros((0,))
        M1 = np.zeros((0,))
        M2 = np.zeros((0,))
        RWC1 = np.zeros((0,))
        RWC2 = np.zeros((0,))
        C1 = np.zeros((0,))
        C2 = np.zeros((0,))
    elif mask is not None and np.where(mask)[0].__len__() == 0:
        corr = np.zeros((n_objects,))
        corr[:] = np.NaN
        overlap = K1 = K2 = M1 = M2 = RWC1 = RWC2 = C1 = C2 = corr
    else:
        lrange = np.arange(n_objects, dtype=np.int32) + 1

        if MeasurementType.CORRELATION in measurement_types:
            corr = measure_correlation_and_slope_from_objects(im1_pixels, im2_pixels, labels, lrange)
            col_order_1 = (im1_name, im2_name, object_name)
            summary += __get_object_result_array(col_order_1, "Correlation coeff", corr)

        if set(measurement_types).intersection({MeasurementType.MANDERS, MeasurementType.RWC, MeasurementType.OVERLAP}) != set():
            # Get channel-specific thresholds from thresholds array
            im1_threshold = im1_thr_percentage
            im2_threshold = im2_thr_percentage
            (
                im1_thr_sum, 
                im2_thr_sum, 
                thr_mask_intersection,
            ) = get_thresholded_images_and_counts (im1_pixels, im2_pixels, im1_threshold, im2_threshold, labels)

            if MeasurementType.MANDERS in measurement_types:
                M1, M2 = measure_manders_coefficient_from_objects(im1_pixels, im2_pixels, im1_thr_sum, im2_thr_sum, thr_mask_intersection,labels, lrange)
                summary += __get_object_result_array((im1_name, im2_name, object_name), "Manders coeff", M1)
                summary += __get_object_result_array((im2_name, im1_name, object_name), "Manders coeff", M2)


            if MeasurementType.RWC in measurement_types:
                RWC1, RWC2 = measure_rwc_coefficient_from_objects(im1_pixels, im2_pixels, im1_thr_sum, im2_thr_sum, thr_mask_intersection,labels, lrange)
                summary += __get_object_result_array((im1_name, im2_name, object_name), "RWC coeff", RWC1)
                summary += __get_object_result_array((im2_name, im1_name, object_name), "RWC coeff", RWC2)


            if MeasurementType.OVERLAP in measurement_types:
                overlap, K1, K2 = measure_overlap_coefficient_from_objects(im1_pixels, im2_pixels, thr_mask_intersection,labels, lrange)
                summary += __get_object_result_array((im1_name, im2_name, object_name), "Overlap coeff", overlap)


        if MeasurementType.COSTES in measurement_types:
            C1, C2 = measure_costes_coefficient_from_objects(
                im1_pixels = im1_pixels, 
                im2_pixels = im2_pixels, 
                im1_costes_pixels = im1_costes_pixels, 
                im2_costes_pixels = im2_costes_pixels, 
                labels = labels, 
                lrange = lrange, 
                im1_scale = im1_scale, 
                im2_scale = im2_scale, 
                costes_method = costes_method,
            )
            summary += __get_object_result_array((im1_name, im2_name, object_name), "Manders coeff (Costes)", C1)
            summary += __get_object_result_array((im2_name, im1_name, object_name), "Manders coeff (Costes)", C2)


    measurements = LibraryMeasurements()
    
    if MeasurementType.CORRELATION in measurement_types:
        measurement = TemplateMeasurementFormat.CORRELATION_FORMAT % (im1_name, im2_name)
        measurements.add_measurement(object_name, measurement, corr)
    if MeasurementType.MANDERS in measurement_types:
        manders_measurement_1 = TemplateMeasurementFormat.MANDERS_FORMAT % (im1_name, im2_name)
        manders_measurement_2 = TemplateMeasurementFormat.MANDERS_FORMAT % (im2_name, im1_name)
        
        measurements.add_measurement(object_name, manders_measurement_1, M1)
        measurements.add_measurement(object_name, manders_measurement_2, M2)
    
    if MeasurementType.RWC in measurement_types:
        rwc_measurement_1 = TemplateMeasurementFormat.RWC_FORMAT % (im1_name, im2_name)
        rwc_measurement_2 = TemplateMeasurementFormat.RWC_FORMAT % (im2_name, im1_name)

        measurements.add_measurement(object_name, rwc_measurement_1, RWC1)
        measurements.add_measurement(object_name, rwc_measurement_2, RWC2)
    
    if MeasurementType.OVERLAP in measurement_types:
        overlap_measurement = TemplateMeasurementFormat.OVERLAP_FORMAT % (im1_name, im2_name)
        k_measurement_1 = TemplateMeasurementFormat.K_FORMAT % (im1_name, im2_name)
        k_measurement_2 = TemplateMeasurementFormat.K_FORMAT % (im2_name, im1_name)
        
        measurements.add_measurement(object_name, overlap_measurement, overlap)
        measurements.add_measurement(object_name, k_measurement_1, K1)
        measurements.add_measurement(object_name, k_measurement_2, K2)
    
    if MeasurementType.COSTES in measurement_types:
        costes_measurement_1 = TemplateMeasurementFormat.COSTES_FORMAT % (im1_name, im2_name)
        costes_measurement_2 = TemplateMeasurementFormat.COSTES_FORMAT % (im2_name, im1_name)
        
        measurements.add_measurement(object_name, costes_measurement_1, C1)
        measurements.add_measurement(object_name, costes_measurement_2, C2)
        col_order_1 = (im1_name, im2_name, object_name)
        summary +=  [
            (*col_order_1, "Mean correlation", "-"),
            (*col_order_1, "Median correlation", "-"),
            (*col_order_1, "Min correlation", "-"),
            (*col_order_1, "Max correlation", "-"),
        ]
        
    return measurements, summary



from ..functions.file_processing import save_object_image_crops, save_object_masks

def savecroppedobjects(
    input_objects,
    save_dir,
    export_as="masks",
    input_image=None,
    file_format="tiff8",
    nested_save=False,
    save_names={"input_filename": None, "input_objects_name": None},
    volumetric=False
    ):
    if export_as.casefold() in ("image", "images"):
        filenames = save_object_image_crops(
            input_image=input_image,
            input_objects=input_objects,
            save_dir=save_dir,
            file_format=file_format,
            nested_save=nested_save,
            save_names=save_names,
            volumetric=volumetric
        )
    elif export_as.casefold() in ("mask", "masks"):
        filenames = save_object_masks(
            input_objects=input_objects,
            save_dir=save_dir,
            file_format=file_format,
            nested_save=nested_save,
            save_names=save_names,
            )
    return filenames



from cellprofiler_library.opts.measureimageoverlap import DM
from cellprofiler_library.functions.measurement import (
    measure_image_overlap_statistics,
    compute_earth_movers_distance,
)


def measureimageoverlap(
    ground_truth_image,
    test_image,
    mask=None,
    calculate_emd=False,
    max_distance=250,
    penalize_missing=False,
    decimation_method: DM = DM.KMEANS,
    max_points=250,
):

    data = measure_image_overlap_statistics(
        ground_truth_image=ground_truth_image, test_image=test_image, mask=mask
    )

    if calculate_emd:
        emd = compute_earth_movers_distance(
            ground_truth_image=ground_truth_image,
            test_image=test_image,
            max_distance=max_distance,
            penalize_missing=penalize_missing,
            decimation_method=decimation_method,
            max_points=max_points,
        )
        data.update({"EarthMoversDistance": emd})
    return data



from pydantic import validate_call, ConfigDict, Field
from typing import Annotated, List, Dict, Any, Tuple
from enum import Enum
import numpy as np

from cellprofiler_library.measurement_model import LibraryMeasurements
from cellprofiler_library.functions.measurement import get_granularity_measurements, ObjectRecord
from cellprofiler_library.functions.image_processing import apply_grayscale_tophat_filter, downsample_image_and_mask
from cellprofiler_library.types import ImageGrayscale, ImageGrayscaleMask

class GranularityMeasurementFormat(str, Enum):
    GRANULARITY = "Granularity_%d_%s"

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def measure_granularity(
        image_name:                 Annotated[str, Field(description="Name of the image")],
        im_pixel_data:              Annotated[ImageGrayscale, Field(description="Pixel values of the image")],
        im_mask:                    Annotated[ImageGrayscaleMask, Field(description="Boolean mask of where the measurements should be made")],
        subsample_size:             Annotated[float, Field(description="Subsampling factor for granularity measurements")],
        image_sample_size:          Annotated[float, Field(description="Subsampling factor for background reduction")],
        element_size:               Annotated[int, Field(description="Radius of structuring element")],
        object_records:             Annotated[List[ObjectRecord], Field(description="Object records")],
        granular_spectrum_length:   Annotated[int, Field(description="Range of the granular spectrum")],
        dimensions:                 Annotated[int, Field(description="Dimensionality of the image")] = 2,
        ) -> Tuple[LibraryMeasurements, List[str]]:
    #
    # Downsample the image and mask
    #
    pixels, mask, new_shape = downsample_image_and_mask(im_pixel_data, im_mask, dimensions, subsample_size)
    
    #
    # Remove background pixels using a greyscale tophat filter
    #
    pixels = apply_grayscale_tophat_filter(pixels, mask, dimensions, image_sample_size, element_size, new_shape)
    
    #
    # Compute measurements
    #
    measurements_arr, image_measurements_arr = get_granularity_measurements(
        im_pixel_data,
        pixels, 
        mask, 
        new_shape,
        granular_spectrum_length,
        dimensions,
        object_records
    )

    measurements = LibraryMeasurements()
    stats_strings = []

    # Process Image Measurements
    for i, gs_value in enumerate(image_measurements_arr):
        spectrum_index = i + 1
        feature_name = GranularityMeasurementFormat.GRANULARITY % (spectrum_index, image_name)
        measurements.add_image_measurement(feature_name, gs_value)
        stats_strings.append("%.2f" % gs_value)

    # Process Object Measurements
    # measurements_arr is list of lists. Outer list index i corresponds to spectrum_index i+1.
    for i, obj_measurements_list in enumerate(measurements_arr):
        spectrum_index = i + 1
        feature_name = GranularityMeasurementFormat.GRANULARITY % (spectrum_index, image_name)
        
        for obj_name, obj_gss in obj_measurements_list:
            measurements.add_measurement(obj_name, feature_name, obj_gss)
            
            # Calculate stats for this object measurement and add to Image measurements
            stat_name_base = f"{feature_name}_{obj_name}"
            
            measurements.add_image_measurement(f"Mean_{stat_name_base}", np.mean(obj_gss) if len(obj_gss) > 0 else 0.0)
            measurements.add_image_measurement(f"Median_{stat_name_base}", np.median(obj_gss) if len(obj_gss) > 0 else 0.0)
            measurements.add_image_measurement(f"Max_{stat_name_base}", np.max(obj_gss) if len(obj_gss) > 0 else 0.0)
            measurements.add_image_measurement(f"Min_{stat_name_base}", np.min(obj_gss) if len(obj_gss) > 0 else 0.0)
            measurements.add_image_measurement(f"StDev_{stat_name_base}", np.std(obj_gss) if len(obj_gss) > 0 else 0.0)

    # Summary for UI display
    summary = [image_name] + stats_strings
    
    return measurements, summary



from typing import Annotated, Optional
from pydantic import Field, validate_call, ConfigDict
from cellprofiler_library.opts.correctilluminationapply import Method
from ..types import Image2D
from ..functions.image_processing import apply_divide, apply_subtract, clip_low, clip_high


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def correct_illumination_apply(
        image_pixels:               Annotated[Image2D, Field(description="Pixel data of image to apply the illumination function to")],
        illum_function_pixel_data:  Annotated[Image2D, Field(description="Pixel data of illumination function")],
        method_divide_or_subtract:  Annotated[Method, Field(description="Method to apply the illumination function")],
        truncate_low:               Annotated[Optional[bool], Field(description="Set output image values less than 0 equal to 0?")],
        truncate_high:              Annotated[Optional[bool], Field(description="Set output image values greater than 1 equal to 1?")],
        ) -> Annotated[Image2D, Field(description="Pixel data of image with illumination function applied")]:
    """
    Perform illumination according to the parameters of one image setting group
    """
    assert image_pixels.shape[:2] == illum_function_pixel_data.shape[:2], "Input image shape and illumination function shape must be equal"
    #
    # Either divide or subtract the illumination image from the original
    #
    if method_divide_or_subtract == Method.DIVIDE:
        output_pixels = apply_divide(image_pixels, illum_function_pixel_data)
    elif method_divide_or_subtract == Method.SUBTRACT:
        output_pixels = apply_subtract(image_pixels, illum_function_pixel_data)
    else:
        raise ValueError(
            "Unhandled option for divide or subtract: %s"
            % method_divide_or_subtract.value
        )
    #
    # Optionally, clip high and low values
    #
    if truncate_low:
        output_pixels = clip_low(output_pixels)
    if truncate_high:
        output_pixels = clip_high(output_pixels)

    return output_pixels



from pydantic import Field, validate_call, ConfigDict
from typing import Optional, Tuple, Annotated, Any, Union

from ..functions.image_processing import (
    get_adaptive_threshold,
    get_global_threshold,
    apply_threshold,
)
from ..opts.threshold import (
    Scope,
    Method,
    Assignment,
    AveragingMethod,
    VarianceMethod,
)
from ..types import ImageGrayscale, ImageGrayscaleMask

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def threshold(
    image:                      Annotated[ImageGrayscale, Field(description="Image to threshold")],
    mask:                       Annotated[Optional[ImageGrayscaleMask], Field(description="Mask to apply to the image")] = None,
    threshold_scope:            Annotated[Scope, Field(description="Thresholding scope")] = Field(default=Scope.GLOBAL),
    threshold_method:           Annotated[Method, Field(description="Thresholding method")] = Field(default=Method.OTSU),
    assign_middle_to_foreground:Annotated[Assignment, Field(description="Assign middle to foreground")] = Field(default=Assignment.FOREGROUND),
    log_transform:              Annotated[bool, Field(description="Log transform")] = Field(default=False),
    threshold_correction_factor:Annotated[float, Field(description="Threshold correction factor")] = Field(default=1),
    threshold_min:              Annotated[Optional[float], Field(description="Minimum threshold")] = Field(default=0),
    threshold_max:              Annotated[Optional[float], Field(description="Maximum threshold")] = Field(default=1),
    window_size:                Annotated[int, Field(description="Window size for adaptive thresholding")] = Field(default=50),
    smoothing:                  Annotated[float, Field(description="Smoothing factor")] = Field(default=0),
    lower_outlier_fraction:     Annotated[float, Field(description="Lower outlier fraction")] = Field(default=0.05),
    upper_outlier_fraction:     Annotated[float, Field(description="Upper outlier fraction")] = Field(default=0.05),
    averaging_method:           Annotated[AveragingMethod, Field(description="Averaging method")] = Field(default=AveragingMethod.MEAN),
    variance_method:            Annotated[VarianceMethod, Field(description="Variance method")] = Field(default=VarianceMethod.STANDARD_DEVIATION),
    number_of_deviations:       Annotated[float, Field(description="Number of deviations")] = Field(default=2.0),
    predefined_threshold:       Annotated[Optional[float], Field(description="Predefined threshold value")] = Field(default=None),
    volumetric:                 Annotated[bool, Field(description="Volumetric thresholding")] = Field(default=False),
    automatic:                  Annotated[bool, Field(description="Automatic thresholding")] = Field(default=False),
    **kwargs:                   Annotated[Any, Field(description="Additional keyword arguments")]
) -> Tuple[
    Annotated[Union[Any, float, int], Field(description="Final threshold")],
    Annotated[Union[Any, float, int], Field(description="Original threshold")],
    Annotated[Union[Any, float, int], Field(description="Guide threshold")],
    Annotated[ImageGrayscaleMask, Field(description="Binary image")],
    Annotated[float, Field(description="Sigma value")],

]:
    """
    Returns three threshold values and a binary image.
    Thresholds returned are:

    Final threshold: Threshold following application of the
    threshold_correction_factor and clipping to min/max threshold

    orig_threshold: The threshold following either adaptive or global
    thresholding strategies, prior to correction

    guide_threshold: Only produced by adaptive threshold, otherwise None.
    This is the global threshold that constrains the adaptive threshold
    within a certain range, as defined by global_limits (default [0.7, 1.5])
    """

    # A predefined threshold has been requested (ie. a manual or measurement one)
    if predefined_threshold is not None:
        final_threshold = predefined_threshold
        final_threshold *= threshold_correction_factor
        # For manual thresholds in the GUI, min/max filtering is not applied
        if threshold_min is not None and threshold_max is not None:
            final_threshold = min(max(final_threshold, threshold_min), threshold_max)
        orig_threshold = predefined_threshold
        guide_threshold = None
        binary_image, sigma = apply_threshold(
            image=image,
            threshold=final_threshold,
            mask=mask,
            smoothing=smoothing
            )
        return final_threshold, orig_threshold, guide_threshold, binary_image, sigma

    if automatic:
        # Use automatic settings
        smoothing = 1
        log_transform = False
        threshold_scope = Scope.GLOBAL
        threshold_method = Method.MINIMUM_CROSS_ENTROPY

    # Only pass robust_background kwargs when selected as the threshold_method
    if threshold_method == Method.ROBUST_BACKGROUND:
        kwargs = {
            "lower_outlier_fraction": lower_outlier_fraction,
            "upper_outlier_fraction": upper_outlier_fraction,
            "averaging_method": averaging_method,
            "variance_method": variance_method,
            "number_of_deviations": number_of_deviations,
        }

    if threshold_scope == Scope.ADAPTIVE:
        final_threshold = get_adaptive_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            window_size=window_size,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            volumetric=volumetric,
            **kwargs,
        )
        orig_threshold = get_adaptive_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            window_size=window_size,
            # If automatic=True, do not correct the threshold
            threshold_min=threshold_min if automatic else 0,
            threshold_max=threshold_max if automatic else 1,
            threshold_correction_factor=threshold_correction_factor if automatic else 1,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            volumetric=volumetric,
            **kwargs,
        )

        guide_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )

        binary_image, sigma = apply_threshold(
            image,
            threshold=final_threshold,
            mask=mask,
            smoothing=smoothing,
        )

        return final_threshold, orig_threshold, guide_threshold, binary_image, sigma

    elif threshold_scope == Scope.GLOBAL:
        final_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            threshold_correction_factor=threshold_correction_factor,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )
        orig_threshold = get_global_threshold(
            image,
            mask=mask,
            threshold_method=threshold_method,
            # If automatic=True, do not correct the threshold
            threshold_min=threshold_min if automatic else 0,
            threshold_max=threshold_max if automatic else 1,
            threshold_correction_factor=threshold_correction_factor if automatic else 1,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            **kwargs,
        )
        guide_threshold = None
        binary_image, sigma = apply_threshold(
            image,
            threshold=final_threshold,
            mask=mask,
            smoothing=smoothing,
        )
        return final_threshold, orig_threshold, guide_threshold, binary_image, sigma



"""Resize module core functionality.

This module contains the extracted resize functions that implement
the core image resizing algorithms.
"""

from typing import Annotated, Optional, Tuple
from pydantic import Field, validate_call, ConfigDict

from cellprofiler_library.types import ImageGrayscaleMask, ImageAny, ImageBinary
from cellprofiler_library.functions.image_processing import apply_resize
from cellprofiler_library.opts.resize import ResizingMethod, DimensionMethod, InterpolationMethod


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def resize_image(
    im_pixel_data:          Annotated[ImageAny, Field(description="Input image pixel data array")],
    im_mask:                Annotated[ImageGrayscaleMask, Field(description="Input image mask")],
    im_dimensions:          Annotated[int, Field(description="Number of spatial dimensions in the image", ge=2, le=3)],
    im_crop_mask:           Annotated[Optional[ImageBinary], Field(description="Image crop mask")],
    size_method:            Annotated[ResizingMethod, Field(description="Method for determining resize dimensions")],
    resizing_factor_x:      Annotated[float, Field(description="X-axis scaling factor", gt=0)],
    resizing_factor_y:      Annotated[float, Field(description="Y-axis scaling factor", gt=0)],
    resizing_factor_z:      Annotated[Optional[float], Field(description="Z-axis scaling factor", gt=0)],
    use_manual_or_image:    Annotated[DimensionMethod, Field(description="Method for specifying dimensions")],
    specific_width:         Annotated[Optional[int], Field(description="Specific width in pixels", ge=1)],
    specific_height:        Annotated[Optional[int], Field(description="Specific height in pixels", ge=1)],
    specific_planes:        Annotated[Optional[int], Field(description="Specific number of planes", ge=1)],
    reference_image_shape:  Annotated[Optional[Tuple[int, ...]], Field(description="Shape of reference image for dimensions")],
    interpolation_method:   Annotated[InterpolationMethod, Field(description="Interpolation method for resizing")],
) -> Tuple[ImageAny, ImageGrayscaleMask, Optional[ImageBinary]]:
    # Apply the resize operation using library
    return apply_resize(
        im_pixel_data,
        im_mask,
        im_dimensions,
        im_crop_mask,
        size_method,
        resizing_factor_x,
        resizing_factor_y,
        resizing_factor_z,
        use_manual_or_image,
        specific_width,
        specific_height,
        specific_planes,
        reference_image_shape,
        interpolation_method,
    )





import numpy
from typing import Callable, Optional, List, Tuple, cast, Union, Annotated
from numpy import isscalar as _isscalar
from pydantic import validate_call, ConfigDict, Field
from cellprofiler_library.opts.imagemath import Operator, BINARY_OUTPUT_OPS, UNARY_INPUT_OPS, BINARY_INPUT_OPS
from cellprofiler_library.types import ImageAny, ImageAnyMask
from cellprofiler_library.functions.image_processing import imagemath_apply_binary_operation, imagemath_apply_unary_operation
isscalar = cast(Callable[[Optional[ImageAny]], bool], _isscalar)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def image_math(
        opval:              Annotated[Operator, Field(description=" Operator to apply to the images or measurements")],
        operands:           Annotated[List[Union[ImageAny, float]], Field(description=" List of images and/or measurements to apply the operator to")],
        masks:              Annotated[Optional[List[Optional[Union[ImageAnyMask, bool]]]], Field(description=" List of masks for the images. Use `True` for measurement operands")],
        output_pixel_data:  Annotated[ImageAny, Field(description=" Output image or measurement")],
        output_mask:        Annotated[Optional[ImageAnyMask], Field(description=" Output mask")],
        ignore_mask:        Annotated[bool, Field(description=" Ignore the masks")],
        image_factors:      Annotated[Optional[List[float]], Field(description=" List of image factors to multiply the images by")],
        exponent:           Annotated[Optional[float], Field(description=" Exponent to raise the images by")],
        after_factor:       Annotated[Optional[float], Field(description=" Multiply the images by this value after the operation")],
        addend:             Annotated[Optional[float], Field(description=" Add this value to the images")],
        truncate_low:       Annotated[Optional[bool], Field(description=" Truncate the images below 0")],
        truncate_high:      Annotated[Optional[bool], Field(description=" Truncate the images above 0")],
        replace_nan:        Annotated[Optional[bool], Field(description=" Replace invalid values with 0")],
        ) -> Tuple[
            ImageAny, Optional[ImageAnyMask]
        ]:
    if opval in BINARY_INPUT_OPS:
        output_pixel_data, output_mask = imagemath_apply_binary_operation(opval, operands, masks, output_pixel_data, output_mask, image_factors, ignore_mask)

    elif opval in UNARY_INPUT_OPS:
        output_pixel_data, output_mask = imagemath_apply_unary_operation(opval, operands, masks, output_pixel_data, output_mask, ignore_mask)

    else:
        raise NotImplementedError(
            "The operation %s has not been implemented" % opval
        )

    # Check to see if there was a measurement & image w/o mask. If so
    # set mask to none
    if isscalar(output_mask):
        output_mask = None
    if opval not in BINARY_OUTPUT_OPS:
        #
        # Post-processing: exponent, multiply, add
        #
        if exponent is not None and exponent != 1:
            output_pixel_data **= exponent
        if after_factor is not None and after_factor != 1:
            output_pixel_data *= after_factor
        if addend is not None and addend != 0:
            output_pixel_data += addend

        #
        # truncate values
        #
        if truncate_low:
            output_pixel_data[output_pixel_data < 0] = 0
        if truncate_high:
            output_pixel_data[output_pixel_data > 1] = 1
        if replace_nan:
            output_pixel_data[numpy.isnan(output_pixel_data)] = 0
    return output_pixel_data, output_mask


from enum import Enum

class Scope(str, Enum):
    GLOBAL = "Global"
    ADAPTIVE = "Adaptive"

class OtsuMethod(str, Enum):
    TWO_CLASS = "Two classes"
    THREE_CLASS = "Three classes"

class Method(str, Enum):
    OTSU = "Otsu"
    MINIMUM_CROSS_ENTROPY = "Minimum Cross-Entropy"
    ROBUST_BACKGROUND = "Robust Background"
    MULTI_OTSU = "Multi-Otsu"
    SAUVOLA = "Sauvola"
    MAX_INTENSITY_PERCENTAGE = "Max Intensity Percentage" # For MeasureColocalization
    MANUAL = "Manual" # For IdentifyPrimaryObjects
    MEASUREMENT = "Measurement" # For IdentifyPrimaryObjects

class Assignment(str, Enum):
    # assign_middle_to_foreground
    FOREGROUND = "Foreground"
    BACKGROUND = "Background"

class AveragingMethod(str, Enum):
    MEAN = "Mean"
    MEDIAN = "Median"
    MODE = "Mode"

class VarianceMethod(str, Enum):
    STANDARD_DEVIATION = "Standard deviation"
    MEDIAN_ABSOLUTE_DEVIATION = "Median absolute deviation"



from ..functions.image_processing import morphological_skeleton_2d, morphological_skeleton_3d

def morphologicalskeleton(image, volumetric):
    if volumetric: 
        return morphological_skeleton_3d(image)
    else:
        return morphological_skeleton_2d(image)




from pydantic import Field, validate_call, ConfigDict
from typing import Annotated, List, Union, Optional

from ..opts.colortogray import ImageChannelType
from ..types import Image2DColor, Image2DGrayscale
from ..functions.image_processing import combine_colortogray, split_hsv, split_rgb, split_multichannel

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def color_to_gray(
          image:              Annotated[Image2DColor, Field(description="Pixel data of image to threshold")],
          image_type:         Annotated[ImageChannelType, Field(description="Image type (RGB, HSV, or Channels)")],
          should_combine:     Annotated[bool, Field(description="Whether to combine or split the image")],
          channels:           Annotated[Optional[List[int]], Field(description="Array of integer identifier for combining")],
          contributions:      Annotated[Optional[List[float]], Field(description="Array of contribution values for combining")],
          ) -> Union[Image2DGrayscale, List[Image2DGrayscale]]:
     if should_combine:
          if channels is None or contributions is None:
               raise ValueError("Must provide channels and contributions when combining")
          return combine_colortogray(image, channels, contributions)
     else:
          return split_colortogray(image, image_type)
     
def split_colortogray(input_image: Image2DColor, image_type:ImageChannelType = ImageChannelType.RGB) -> List[Image2DGrayscale]:
     if image_type == ImageChannelType.RGB:
          return split_rgb(input_image) 
     elif image_type == ImageChannelType.HSV:
          return split_hsv(input_image)
     elif image_type == ImageChannelType.CHANNELS:
          return split_multichannel(input_image)
     else:
          raise ValueError(f"Unsupported image type: {image_type}")



from centrosome import cpmorphology
import numpy
import scipy
from typing import Annotated, Optional, Union, Tuple
from cellprofiler_library.modules import threshold
from cellprofiler_library.functions.object_processing import (
    separate_neighboring_objects,
    filter_on_border,
    filter_on_size
)
from pydantic import validate_call, ConfigDict, Field
from cellprofiler_library.types import Image2DGrayscale, Image2DGrayscaleMask, ObjectSegmentation, ImageGrayscale, ImageGrayscaleMask
import cellprofiler_library.opts.threshold as ThresholdOpts
from cellprofiler_library.opts.identifyprimaryobjects import UnclumpMethod, WatershedMethod, FillHolesMethod

CP_Output_Type = Tuple[
    ObjectSegmentation,
    ObjectSegmentation,
    ObjectSegmentation,
    ObjectSegmentation,
    ObjectSegmentation,
    ObjectSegmentation,
    float,
    int,
    Union[float, ImageGrayscale],
    Union[float, ImageGrayscale],
    Optional[float],
    ImageGrayscaleMask,
    float,
    float
]


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def identifyprimaryobjects(
    image:                      Annotated[Image2DGrayscale, Field(description="Input image")],
    mask:                       Annotated[Optional[Image2DGrayscaleMask], Field(description="Input mask")] = None,      
    threshold_method:           Annotated[ThresholdOpts.Method, Field(description="Thresholding method")] = ThresholdOpts.Method.MINIMUM_CROSS_ENTROPY,
    threshold_scope:            Annotated[ThresholdOpts.Scope, Field(description="Thresholding scope")] = ThresholdOpts.Scope.GLOBAL,
    assign_middle_to_foreground:Annotated[ThresholdOpts.Assignment, Field(description="Assign middle to foreground")] = ThresholdOpts.Assignment.BACKGROUND,
    log_transform:              Annotated[bool, Field(description="Apply log transform to image before thresholding")] = False,
    threshold_correction_factor:Annotated[float, Field(description="Multiply threshold by this factor")] = 1.0,
    threshold_min:              Annotated[float, Field(description="Minimum threshold value")] = 0.0,
    threshold_max:              Annotated[float, Field(description="Maximum threshold value")] = 1.0,
    window_size:                Annotated[int, Field(description="Size of window for thresholding")] = 50,
    threshold_smoothing:        Annotated[float, Field(description="Smoothing factor for thresholding")] = 0.0,
    lower_outlier_fraction:     Annotated[float, Field(description="Fraction of pixels to use for lower outlier detection")] = 0.05,
    upper_outlier_fraction:     Annotated[float, Field(description="Fraction of pixels to use for upper outlier detection")] = 0.05,
    averaging_method:           Annotated[ThresholdOpts.AveragingMethod, Field(description="Averaging method for thresholding")] = ThresholdOpts.AveragingMethod.MEAN,
    variance_method:            Annotated[ThresholdOpts.VarianceMethod, Field(description="Variance method for thresholding")] = ThresholdOpts.VarianceMethod.STANDARD_DEVIATION,
    number_of_deviations:       Annotated[float, Field(description="Number of deviations for thresholding")] = 2.0,
    automatic:                  Annotated[bool, Field(description="Automatically determine thresholding parameters")] = False,
    exclude_size:               Annotated[bool, Field(description="Exclude objects smaller than this size")] = True,
    min_size:                   Annotated[int, Field(description="Minimum object size")] = 10,
    max_size:                   Annotated[int, Field(description="Maximum object size")] = 40,
    exclude_border_objects:     Annotated[bool, Field(description="Exclude objects touching the border")] = True,
    unclump_method:             Annotated[UnclumpMethod, Field(description="Unclump method for thresholding")] = UnclumpMethod.INTENSITY,
    watershed_method:           Annotated[WatershedMethod, Field(description="Watershed method for thresholding")] = WatershedMethod.INTENSITY,
    fill_holes_method:          Annotated[FillHolesMethod, Field(description="Fill holes method for thresholding")] = FillHolesMethod.THRESHOLDING,
    smoothing_filter_size:      Annotated[Optional[int], Field(description="Smoothing filter size for thresholding")] = None,
    automatic_suppression:      Annotated[bool, Field(description="Automatically calculate suppression size")] = True,
    maxima_suppression_size:    Annotated[float, Field(description="Suppression size for thresholding")] = 7.0,
    low_res_maxima:             Annotated[bool, Field(description="Use low-resolution maxima for thresholding")] = True,
    maximum_object_count:       Annotated[Optional[int], Field(description="Maximum number of objects to threshold")] = None,
    predefined_threshold:       Annotated[Optional[float], Field(description="Predefined threshold value")] = None,
    return_cp_output:           Annotated[bool, Field(description="Return CellProfiler output")] = False
) -> Union[ObjectSegmentation, CP_Output_Type]:
    
    if automatic:
        # Define automatic settings by recursively calling identifyprimaryobjects and defining 
        # the automatic settings
        return identifyprimaryobjects(
            image,
            mask=mask,
            threshold_smoothing=1,
            log_transform=False,
            threshold_scope=ThresholdOpts.Scope.GLOBAL,
            threshold_method=ThresholdOpts.Method.MINIMUM_CROSS_ENTROPY,
            automatic=False,  # False, since this recursive call defines what the automatic settings are
            exclude_size=exclude_size,
            min_size=min_size,
            max_size=max_size,
            exclude_border_objects=exclude_border_objects,
            unclump_method=UnclumpMethod.INTENSITY,
            watershed_method=WatershedMethod.INTENSITY,
            fill_holes_method=FillHolesMethod.THRESHOLDING,
            smoothing_filter_size=None,
            low_res_maxima=True if min_size > 10 else False,
            automatic_suppression=True,
            return_cp_output=return_cp_output,
        )

    final_threshold, orig_threshold, guide_threshold, binary_image, sigma = threshold(
        image=image,
        mask=mask,
        threshold_scope=threshold_scope,
        threshold_method=threshold_method,
        assign_middle_to_foreground=assign_middle_to_foreground,
        log_transform=log_transform,
        threshold_correction_factor=threshold_correction_factor,
        threshold_min=threshold_min,
        threshold_max=threshold_max,
        window_size=window_size,
        smoothing=threshold_smoothing,
        lower_outlier_fraction=lower_outlier_fraction,
        upper_outlier_fraction=upper_outlier_fraction,
        averaging_method=averaging_method,
        variance_method=variance_method,
        number_of_deviations=number_of_deviations,
        volumetric=False,  # IdentifyPrimaryObjects does not support 3D
        predefined_threshold=predefined_threshold,
    )

    global_threshold = numpy.mean(numpy.atleast_1d(final_threshold))

    if fill_holes_method == FillHolesMethod.THRESHOLDING.value:
        binary_image = cpmorphology.fill_labeled_holes(
            binary_image, size_fn=lambda size, is_foreground: size < max_size * max_size
        )

    # Label the thresholded image
    labeled_image = scipy.ndimage.label(binary_image, numpy.ones((3, 3), bool))[0]

    if smoothing_filter_size is None:
        declump_smoothing_filter_size = 2.35 * min_size / 3.5
    else:
        declump_smoothing_filter_size = smoothing_filter_size

    # If no declumping is selected, a maxima image is not returned
    if return_cp_output:
        (
            labeled_image,
            labeled_maxima,
            maxima_suppression_size,
        ) = separate_neighboring_objects(
            image,
            labeled_image=labeled_image,
            mask=mask,
            unclump_method=unclump_method,
            watershed_method=watershed_method,
            fill_holes_method=fill_holes_method,
            filter_size=declump_smoothing_filter_size,
            min_size=min_size,
            max_size=max_size,
            low_res_maxima=low_res_maxima,
            automatic_suppression=automatic_suppression,
            maxima_suppression_size=maxima_suppression_size,
            return_cp_output=True,
        )
    # Maxima image will be retuened
    else:
        labeled_image = separate_neighboring_objects(
            image,
            labeled_image=labeled_image,
            mask=mask,
            unclump_method=unclump_method,
            watershed_method=watershed_method,
            fill_holes_method=fill_holes_method,
            filter_size=declump_smoothing_filter_size,
            min_size=min_size,
            max_size=max_size,
            low_res_maxima=low_res_maxima,
            maxima_suppression_size=maxima_suppression_size,
            automatic_suppression=automatic_suppression,
        )

    unedited_labels = labeled_image.copy()

    # Filter out objects touching the border or mask
    border_excluded_labeled_image = labeled_image.copy()
    if exclude_border_objects:
        labeled_image = filter_on_border(labeled_image, mask)
    border_excluded_labeled_image[labeled_image > 0] = 0

    # Filter out small and large objects
    size_excluded_labeled_image = labeled_image.copy()

    # If requested, remove cells that are outside the size range
    if exclude_size:
        labeled_image, small_removed_labels = filter_on_size(
            labeled_image, min_size, max_size, return_only_small=True
        )
    else:
        labeled_image, small_removed_labels = labeled_image, labeled_image.copy()

    size_excluded_labeled_image[labeled_image > 0] = 0

    #
    # Fill holes again after watershed
    #
    if fill_holes_method != FillHolesMethod.NEVER:
        labeled_image = cpmorphology.fill_labeled_holes(labeled_image)

    # Relabel the image
    labeled_image, object_count = cpmorphology.relabel(labeled_image)

    if maximum_object_count is not None:
        if object_count > maximum_object_count:
            labeled_image = numpy.zeros(labeled_image.shape, int)
            border_excluded_labeled_image = numpy.zeros(labeled_image.shape, int)
            size_excluded_labeled_image = numpy.zeros(labeled_image.shape, int)
            object_count = 0
            
    if return_cp_output:
        return (
            labeled_image,
            unedited_labels,
            small_removed_labels,
            size_excluded_labeled_image,
            border_excluded_labeled_image,
            labeled_maxima,
            maxima_suppression_size,
            object_count,
            final_threshold,
            orig_threshold,
            guide_threshold,
            binary_image,
            global_threshold,
            sigma,
        )
    else:
        return labeled_image




"""
Options and enums for ShrinkToObjectCenters module
"""

# Note: ShrinkToObjectCenters is a simple morphological operation module that doesn't require
# complex enums. It has no settings beyond basic "Object Processing".



from typing import Tuple
import numpy
import skimage
import scipy

import centrosome
import centrosome.zernike

from cellprofiler_library.functions.measurement import measure_object_size_shape
from cellprofiler_library.opts.objectsizeshapefeatures import ObjectSizeShapeFeatures
from cellprofiler_library.functions.segmentation import (
    _validate_dense,
    convert_dense_to_label_set,
)

DEFAULT_INVALID_VALUE_DTYPE = {
    numpy.float64: numpy.nan,
    numpy.float32: numpy.nan,
    numpy.float16: numpy.nan,
    numpy.uint8: 0,
    numpy.uint16: 0,
    numpy.uint32: 0,
    numpy.uint64: 0,
    numpy.int8: 0,
    numpy.int16: 0,
    numpy.int32: 0,
    numpy.int64: 0,
    numpy.bool_: False,
    numpy.object_: None,
    numpy.str_: "",
}

def measureobjectsizeshape(
    objects,
    calculate_advanced: bool = True,
    calculate_zernikes: bool = True,
    volumetric: bool = False,
    spacing: Tuple = None,
):
    """
    Objects: dense, sparse, ijv, or label objects?
    For now, we will assume dense
    """
    # _validate_dense(objects)

    # Define the feature names
    feature_names = list(ObjectSizeShapeFeatures.F_STANDARD.value)
    if volumetric:
        feature_names += list(ObjectSizeShapeFeatures.F_STD_3D.value)
        if calculate_advanced:
            feature_names += list(ObjectSizeShapeFeatures.F_ADV_3D.value)
    else:
        feature_names += list(ObjectSizeShapeFeatures.F_STD_2D.value)
        if calculate_zernikes:
            feature_names += [
                f"Zernike_{index[0]}_{index[1]}"
                for index in centrosome.zernike.get_zernike_indexes(
                    ObjectSizeShapeFeatures.ZERNIKE_N.value + 1
                )
            ]
        if calculate_advanced:
            feature_names += list(ObjectSizeShapeFeatures.F_ADV_2D.value)

    if len(objects[objects != 0]) == 0:
        data = dict(zip(feature_names, [None] * len(feature_names)))
        for ft in feature_names:
            data[ft] = numpy.zeros((0,))
        return data

    if not volumetric:
        desired_properties = [
            "label",
            "image",
            "area",
            "perimeter",
            "bbox",
            "bbox_area",
            "major_axis_length",
            "minor_axis_length",
            "orientation",
            "centroid",
            "equivalent_diameter",
            "extent",
            "eccentricity",
            "convex_area",
            "solidity",
            "euler_number",
        ]
        if calculate_advanced:
            desired_properties += [
                "inertia_tensor",
                "inertia_tensor_eigvals",
                "moments",
                "moments_central",
                "moments_hu",
                "moments_normalized",
            ]
    else:
        desired_properties = [
            "label",
            "image",
            "area",
            "centroid",
            "bbox",
            "bbox_area",
            "major_axis_length",
            "minor_axis_length",
            "extent",
            "equivalent_diameter",
            "euler_number",
        ]
        if calculate_advanced:
            desired_properties += [
                "solidity",
            ]

    labels = convert_dense_to_label_set(objects, validate=False)
    labels = [i[0] for i in labels]  # Just need the labelmaps, not indices

    if len(labels) > 1:
        # Overlapping labels
        features_to_record = {}
        for labelmap in labels:
            buffer, measured_labels, nobjects = measure_object_size_shape(
                labels=labelmap,
                desired_properties=desired_properties,
                calculate_zernikes=calculate_zernikes,
                calculate_advanced=calculate_advanced,
                spacing=spacing,
            )
            for f, m in buffer.items():
                if f in features_to_record:
                    features_to_record[f] = numpy.concatenate(
                        (features_to_record[f], m)
                    )
                else:
                    features_to_record[f] = m
    else:
        features_to_record, measured_labels, nobjects = measure_object_size_shape(
            labels=labels[0],
            desired_properties=desired_properties,
            calculate_zernikes=calculate_zernikes,
            calculate_advanced=calculate_advanced,
            spacing=spacing,
        )

    # ensure that all objects (objects.indices) are represented in the
    # output, even if they are not present in the label matrix. Fill with nan if missing
    if len(measured_labels) < nobjects:
        for i in objects.indices:
            if i not in measured_labels:
                for f in features_to_record:
                    features_to_record[f] = numpy.insert(
                        features_to_record[f], i-1, DEFAULT_INVALID_VALUE_DTYPE.get(
                            features_to_record[f].dtype.type, numpy.nan
                        )
                    )


    return features_to_record



"""Resize module enums and configuration options."""

from enum import Enum


class ResizingMethod(str, Enum):
    """Method for specifying resize dimensions."""
    BY_FACTOR = "Resize by a fraction or multiple of the original size"
    TO_SIZE = "Resize by specifying desired final dimensions"


class DimensionMethod(str, Enum):
    """Method for specifying target dimensions when resizing to size."""
    MANUAL = "Manual"
    IMAGE = "Image"


class InterpolationMethod(str, Enum):
    """Interpolation method for resize operation."""
    NEAREST_NEIGHBOR = "Nearest Neighbor"
    BILINEAR = "Bilinear"
    BICUBIC = "Bicubic"



from typing import Annotated, Optional, Tuple, List
from pydantic import Field, validate_call, ConfigDict

from ..types import ImageAny, ObjectLabelSet, ImageColor
from ..opts.overlayoutlines import BrightnessMode, LineMode
from ..functions.image_processing import (
    create_overlay_base_image,
    overlay_outlines_grayscale,
    overlay_outlines_color,
)


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def overlay_outlines(
    brightness_mode:    Annotated[BrightnessMode, Field(description="Brightness control mode for grayscale outline processing", default=BrightnessMode.MAX_POSSIBLE)], 
    line_mode:          Annotated[LineMode, Field(description="Line style mode for drawing outlines", default=LineMode.INNER)],
    obj_shape:          Annotated[Optional[Tuple[int, ...]], Field(description="Object dimensions for creating blank base image when no input image provided", default=None)], 
    obj_dimensions:     Annotated[Optional[int], Field(description="Number of dimensions for objects (2D or 3D) when creating blank image", default=None)], 
    im_pixel_data:      Annotated[Optional[ImageAny], Field(description="Input image pixel data to overlay outlines on (None to create blank image)", default=None)], 
    im_multichannel:    Annotated[bool, Field(description="Whether input image has multiple color channels (False if overlaying on blank image)", default=False)], 
    im_dimensions:      Annotated[Optional[int], Field(description="Number of spatial dimensions in input image (2D or 3D), None if overlaying on blank image", default=None)],
    object_labels_list: Annotated[List[ObjectLabelSet], Field(description="List of object label sets containing segmented object data for outline generation", default_factory=list)],
    colors_list:        Annotated[Optional[List[Tuple[int, int, int]]], Field(description="List of RGB color tuples (0-255) for colored outline mode (None for grayscale)", default=None)],
    is_volumetric:      Annotated[bool, Field(description="Whether objects are 3D volumetric data requiring plane-wise processing", default=False)]
) -> Tuple[ImageAny, ImageColor, Optional[int]]:
    """Dispatcher function that centralizes all overlay outlines business logic.
    
    This function orchestrates the overlay outlines workflow by creating a base image
    (either blank or from input) and then applying either color or grayscale outline
    processing based on the presence of a colors list.
    
    Args:
        brightness_mode: Brightness control mode determining outline intensity in grayscale mode
        line_mode: Line style mode controlling how outlines are drawn (inner, outer, thick)
        obj_shape: Object dimensions for creating blank base image when no input provided
        obj_dimensions: Spatial dimensionality (2 or 3) for objects when creating blank image
        im_pixel_data: Input image data to overlay outlines on (None creates blank image)
        im_multichannel: Whether input image contains multiple color channels (False if overlaying on blank image)
        im_dimensions: Spatial dimensionality (2 or 3) of input image, None if overlaying on blank image
        object_labels_list: Object label sets containing segmented regions for outline generation
        colors_list: RGB color values for colored outlines (None triggers grayscale mode)
        is_volumetric: Whether objects require 3D plane-wise processing
        
    Returns:
        Tuple of (processed_pixel_data, base_image, dimensions) for the final result
    """
    # Create base image (blank or from input)
    base_image, dimensions = create_overlay_base_image(
        obj_shape, obj_dimensions, 
        im_pixel_data, im_multichannel, im_dimensions
    )
    
    if colors_list is not None:
        # Color outline processing
        pixel_data = overlay_outlines_color(base_image.copy(), object_labels_list, colors_list, line_mode, is_volumetric)
    else:
        # Black and white outline processing
        pixel_data = overlay_outlines_grayscale(
            base_image, brightness_mode, 
            object_labels_list, line_mode, is_volumetric
        )
        
    return pixel_data, base_image, dimensions



import numpy
from numpy.typing import NDArray
from typing import List, Tuple, Optional
from pydantic import validate_call, ConfigDict
from cellprofiler_library.opts.graytocolor import Scheme
from cellprofiler_library.types import Image2DGrayscale, Image2DColor
from cellprofiler_library.functions.image_processing import gray_to_rgb, gray_to_cmyk, gray_to_composite_color, gray_to_stacked_color

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def gray_to_color(
    pixel_data_arr: List[Optional[Image2DGrayscale]],
    scheme: Scheme,
    color_array: Optional[List[Tuple[int, int, int]]] = None,
    weight_array: Optional[List[float]] = None,
    adjustment_factor_array: Optional[List[float]] = None,
    intensities: List[Tuple[float, ...]]=[(1.0, 0.0, 0.0), (0.0, 1.0, 0.0), (0.0, 0.0, 1.0)],
    wants_rescale: bool=True,
) -> Image2DColor:
    rgb_pixel_data = None
    if scheme == Scheme.RGB:
        assert adjustment_factor_array is not None, "adjustment_factor_array must be provided for RGB mode"
        assert intensities is not None, "intensities must be provided for RGB mode"
        rgb_pixel_data = gray_to_rgb(
            pixel_data_arr = pixel_data_arr,
            adjustment_factor_array = adjustment_factor_array,
            intensities = intensities,
            wants_rescale = wants_rescale,
        )
    elif scheme == Scheme.CMYK:
        assert adjustment_factor_array is not None, "adjustment_factor_array must be provided for CMYK mode"
        assert intensities is not None, "intensities must be provided for CMYK mode"
        rgb_pixel_data = gray_to_cmyk(
            pixel_data_arr = pixel_data_arr,
            adjustment_factor_array = adjustment_factor_array,
            intensities = intensities,
            wants_rescale = wants_rescale,
        )
    elif scheme == Scheme.COMPOSITE:
        assert color_array is not None, "color_array must be provided for composite mode"
        assert weight_array is not None, "weight_array must be provided for composite mode"
        rgb_pixel_data = gray_to_composite_color(
            pixel_data_arr = pixel_data_arr,
            color_array = color_array,
            weight_array = weight_array,
            wants_rescale = wants_rescale,
        )
    elif scheme == Scheme.STACK:
        assert pixel_data_arr is not None, "pixel_data_arr must be provided for stack mode"
        rgb_pixel_data = gray_to_stacked_color(
            pixel_data_arr = pixel_data_arr,
        )
    else:
        raise ValueError(f"Unimplemented scheme: {scheme}")
    if scheme.value != Scheme.STACK and wants_rescale:
        # If we rescaled, clip values that went out of range after multiplication
        rgb_pixel_data[rgb_pixel_data > 1] = 1
    return rgb_pixel_data




import numpy
import skimage
import scipy
import centrosome
import centrosome.propagate
from typing import Annotated, Optional, Union, Tuple
from pydantic import validate_call, ConfigDict, Field
from cellprofiler_library.types import Image2DGrayscale, Image2DGrayscaleMask, ObjectSegmentation, ImageGrayscale, ImageGrayscaleMask
from cellprofiler_library.modules import threshold
from cellprofiler_library.functions.object_processing import filter_labels
from cellprofiler_library.opts.identifysecondaryobjects import SecondaryObjectMethod
import cellprofiler_library.opts.threshold as ThresholdOpts


CP_Output_Type = Tuple[
    Union[float, ImageGrayscale],
    Union[float, ImageGrayscale],
    Optional[float],
    ImageGrayscaleMask,
    float,
    ObjectSegmentation,
    ObjectSegmentation,
    ObjectSegmentation
]

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def identifysecondaryobjects(
    image:                      Annotated[Image2DGrayscale, Field(description="Input image")],
    objects:                    Annotated[ObjectSegmentation, Field(description="Object segmentations")],
    unedited_objects:           Annotated[Optional[ObjectSegmentation], Field(description="Unedited object segmentations")] = None,
    mask:                       Annotated[Optional[Image2DGrayscaleMask], Field(description="Input mask")] = None,
    secondary_object_method:    Annotated[SecondaryObjectMethod , Field(description="Method to determine edges of secondary objects")] = SecondaryObjectMethod.PROPAGATION,
    threshold_method:           Annotated[ThresholdOpts.Method, Field(description="Thresholding method")] = ThresholdOpts.Method.MINIMUM_CROSS_ENTROPY, 
    threshold_scope:            Annotated[ThresholdOpts.Scope, Field(description="Thresholding scope")] = ThresholdOpts.Scope.GLOBAL, 
    assign_middle_to_foreground:Annotated[ThresholdOpts.Assignment, Field(description="Assign middle to foreground")] = ThresholdOpts.Assignment.BACKGROUND, 
    log_transform:              Annotated[bool, Field(description="Apply log transform to image before thresholding")] = False,
    threshold_correction_factor:Annotated[float, Field(description="Multiply threshold by this factor")] = 1.0,
    threshold_min:              Annotated[float, Field(description="Minimum threshold value")] = 0.0,
    threshold_max:              Annotated[float, Field(description="Maximum threshold value")] = 1.0,
    window_size:                Annotated[int, Field(description="Size of window for thresholding")] = 50,
    threshold_smoothing:        Annotated[float, Field(description="Smoothing factor for thresholding")] = 0.0,
    lower_outlier_fraction:     Annotated[float, Field(description="Fraction of pixels to use for lower outlier detection")] = 0.05,
    upper_outlier_fraction:     Annotated[float, Field(description="Fraction of pixels to use for upper outlier detection")] = 0.05,
    averaging_method:           Annotated[ThresholdOpts.AveragingMethod, Field(description="Averaging method for thresholding")] = ThresholdOpts.AveragingMethod.MEAN, 
    variance_method:            Annotated[ThresholdOpts.VarianceMethod, Field(description="Variance method for thresholding")] = ThresholdOpts.VarianceMethod.STANDARD_DEVIATION, 
    number_of_deviations:       Annotated[float, Field(description="Number of deviations for thresholding")] = 2.0,
    predefined_threshold:       Annotated[Optional[float], Field(description="Predefined threshold value")] = None,
    distance_to_dilate:         Annotated[int, Field(description="Number of pixels by which to expand the primary objects")] = 10,
    fill_holes:                 Annotated[bool, Field(description="Fill holes in identified objects?")] = True,
    discard_edge:               Annotated[bool, Field(description="Discard objects touching the edge of the image")] = False,
    regularization_factor:      Annotated[float, Field(description="Regularization factor")] = 0.05,
    return_cp_output:           Annotated[bool, Field(description="Return CellProfiler output")] = False,
    ) -> Union[ObjectSegmentation, CP_Output_Type]:
    if image.shape != objects.shape:
        raise ValueError(
            f"""
            The input image shape {image.shape} does not match the input objects
            shape {objects.shape} If they are paired correctly you may want to
            use the Resize, ResizeObjects or Crop module(s) to make them the
            same size.
        """
        )

    if secondary_object_method != SecondaryObjectMethod.DISTANCE_N:
        (
            final_threshold,
            orig_threshold,
            guide_threshold,
            binary_image, # thresholded_image
            sigma,
        ) = threshold(
            image=image,
            mask=mask,
            threshold_scope=threshold_scope,
            threshold_method=threshold_method,
            assign_middle_to_foreground=assign_middle_to_foreground,
            log_transform=log_transform,
            threshold_correction_factor=threshold_correction_factor,
            threshold_min=threshold_min,
            threshold_max=threshold_max,
            window_size=window_size,
            smoothing=threshold_smoothing,
            lower_outlier_fraction=lower_outlier_fraction,
            upper_outlier_fraction=upper_outlier_fraction,
            averaging_method=averaging_method,
            variance_method=variance_method,
            number_of_deviations=number_of_deviations,
            volumetric=False,  # IDSecondary does not support 3D
            predefined_threshold=predefined_threshold,
        )
    else:
        final_threshold = None
        orig_threshold = None
        guide_threshold = None
        binary_image = None
        sigma = None

    # Input primary objects that will be altered on edge discard
    segmented_labels = objects

    #
    # Get the following labels:
    # * all edited labels
    # * labels touching the edge, including small removed
    #
    if unedited_objects is None:
        labels_in = objects
    else:
        labels_in = unedited_objects.copy()
    labels_touching_edge = numpy.hstack(
        (labels_in[0, :], labels_in[-1, :], labels_in[:, 0], labels_in[:, -1])
    )
    labels_touching_edge = numpy.unique(labels_touching_edge)
    is_touching = numpy.zeros(numpy.max(labels_in) + 1, bool)
    is_touching[labels_touching_edge] = True
    is_touching = is_touching[labels_in]

    labels_in[(~is_touching) & (objects == 0)] = 0
    #
    # Stretch the input labels to match the image size. If there's no
    # label matrix, then there's no label in that area.
    #
    if tuple(labels_in.shape) != tuple(image.shape):
        tmp = numpy.zeros(image.shape, labels_in.dtype)
        i_max = min(image.shape[0], labels_in.shape[0])
        j_max = min(image.shape[1], labels_in.shape[1])
        tmp[:i_max, :j_max] = labels_in[:i_max, :j_max]
        labels_in = tmp

    if secondary_object_method in (SecondaryObjectMethod.DISTANCE_B, SecondaryObjectMethod.DISTANCE_N):
        if secondary_object_method == SecondaryObjectMethod.DISTANCE_N:
            distances, (i, j) = scipy.ndimage.distance_transform_edt(
                labels_in == 0, return_indices=True
            )
            labels_out = numpy.zeros(labels_in.shape, int)
            dilate_mask = distances <= distance_to_dilate
            labels_out[dilate_mask] = labels_in[i[dilate_mask], j[dilate_mask]] 
        else:
            labels_out, distances = centrosome.propagate.propagate(
                image, labels_in, binary_image, 1.0
            )
            labels_out[distances > distance_to_dilate] = 0
            labels_out[labels_in > 0] = labels_in[labels_in > 0]
        if fill_holes:
            label_mask = labels_out == 0
            small_removed_segmented_out = centrosome.cpmorphology.fill_labeled_holes(
                labels_out, mask=label_mask
            )
        else:
            small_removed_segmented_out = labels_out
        #
        # Create the final output labels by removing labels in the
        # output matrix that are missing from the segmented image
        #

        segmented_out = filter_labels(
            small_removed_segmented_out, objects, mask=mask, discard_edge=discard_edge
        )
    elif secondary_object_method == SecondaryObjectMethod.PROPAGATION:
        labels_out, distance = centrosome.propagate.propagate(
            image, labels_in, binary_image, regularization_factor
        )
        if fill_holes:
            label_mask = labels_out == 0
            small_removed_segmented_out = centrosome.cpmorphology.fill_labeled_holes(
                labels_out, mask=label_mask
            )
        else:
            small_removed_segmented_out = labels_out.copy()
        segmented_out = filter_labels(
            small_removed_segmented_out, objects, mask=mask, discard_edge=discard_edge
        )
    elif secondary_object_method == SecondaryObjectMethod.WATERSHED_GRADIENT:
        #
        # First, apply the sobel filter to the image (both horizontal
        # and vertical). The filter measures gradient.
        #
        sobel_image = numpy.abs(scipy.ndimage.sobel(image))
        #
        # Combine the image mask and threshold to mask the watershed
        #
        watershed_mask = numpy.logical_or(binary_image, labels_in > 0)
        watershed_mask = numpy.logical_and(watershed_mask, mask)

        #
        # Perform the first watershed
        #
        labels_out = skimage.segmentation.watershed(
            connectivity=numpy.ones((3, 3), bool),
            image=sobel_image,
            markers=labels_in,
            mask=watershed_mask,
        )
        if fill_holes:
            label_mask = labels_out == 0
            small_removed_segmented_out = centrosome.cpmorphology.fill_labeled_holes(
                labels_out, mask=label_mask
            )
        else:
            small_removed_segmented_out = labels_out.copy()
        segmented_out = filter_labels(
            small_removed_segmented_out, objects, mask=mask, discard_edge=discard_edge
        )
    elif secondary_object_method == SecondaryObjectMethod.WATERSHED_IMAGE:
        #
        # invert the image so that the maxima are filled first
        # and the cells compete over what's close to the threshold
        #
        inverted_img = 1 - image
        #
        # Same as above, but perform the watershed on the original image
        #
        watershed_mask = numpy.logical_or(binary_image, labels_in > 0)
        watershed_mask = numpy.logical_and(watershed_mask, mask)
        #
        # Perform the watershed
        #
        labels_out = skimage.segmentation.watershed(
            connectivity=numpy.ones((3, 3), bool),
            image=inverted_img,
            markers=labels_in,
            mask=watershed_mask,
        )
        if fill_holes:
            label_mask = labels_out == 0
            small_removed_segmented_out = centrosome.cpmorphology.fill_labeled_holes(
                labels_out, mask=label_mask
            )
        else:
            small_removed_segmented_out = labels_out
        segmented_out = filter_labels(
            small_removed_segmented_out, objects, mask=mask, discard_edge=discard_edge
        )
    else:
        raise ValueError(f"Method {secondary_object_method} is not supported")
    
    if discard_edge:
        lookup = scipy.ndimage.maximum(
            segmented_out,
            objects,
            list(range(numpy.max(objects) + 1)),
        )
        lookup = centrosome.cpmorphology.fixup_scipy_ndimage_result(lookup)
        lookup[0] = 0
        lookup[lookup != 0] = numpy.arange(numpy.sum(lookup != 0)) + 1
        segmented_labels = lookup[objects]
        segmented_out = lookup[segmented_out]

    if return_cp_output:
        return (
            final_threshold,
            orig_threshold,
            guide_threshold,
            binary_image,
            sigma, 
            # Input primary objects, but if discard_edge==True, will be primary
            # objects that have been discarded due to edge touching secondary
            # objects following edge discard
            segmented_labels,
            segmented_out, # The actual output objects
            small_removed_segmented_out
        )
    else:
        return segmented_out



from enum import Enum

class StructuringElementShape2D(str, Enum):
    DIAMOND = "Diamond"
    DISK = "Disk"
    SQUARE = "Square"
    STAR = "Star"
    
class StructuringElementShape3D(str, Enum):
    BALL = "Ball"
    CUBE = "Cube"
    OCTAHEDRON = "Octahedron"


from typing import Any, Iterable, Optional, Union, Annotated, get_origin, get_args, Sequence, Tuple
from pydantic import Field, AfterValidator
import numpy as np
import numpy.typing as npt
from numpy.typing import NDArray


class CellProfilerInputValidationError(ValueError):
    def __init__(self, message: str):
        self.message = f"CellProfiler Input Validation Error: {message}"
        super().__init__(self.message)

    def __str__(self):
        return self.message
    
""" 
Image Types can be classified into:
3D/2D: "3D" or "2D"
Single channel/multi-channel: "single" or "multi"
Data type: uint8, uint16, float32, float64, bool
Tiled or single: "tiled" or "single"
CellProfiler expects z to be the first axis and c to be the last. Refer to reader implementations for details.
"""
def create_type_validator(is_3d: bool, is_multi_channel: bool, is_tiled: bool, dtype: type, shape_override: Optional[Iterable[int]] = None):
    def validate(input_image: npt.NDArray[Any]):
        if shape_override is not None:
            if input_image.shape != shape_override:
                raise CellProfilerInputValidationError(f"Expected an array of shape {shape_override}, got {input_image.shape}") 
        else:
            if is_3d:
                if input_image.ndim < 3 or input_image.ndim > 4: # not 3D or not 3D multi-channel
                    raise CellProfilerInputValidationError(f"Expected a 3D or 4D array (z, y, x, [c]), got {input_image.ndim}D")
                if input_image.ndim == 3 and input_image.shape[0] <= 1: # Only one z channel
                    raise CellProfilerInputValidationError(f"Expected a 3D array with at least 2 z channels, got {input_image.shape[1]}")
                if input_image.ndim == 4:
                    if input_image.shape[1] <= 1:
                        raise CellProfilerInputValidationError(f"Expected a 4D array with at least 2 z channels, got {input_image.shape[2]}")
                    if input_image.shape[0] <= 1:
                        raise CellProfilerInputValidationError(f"Expected a 4D array with at least 2 c channels, got {input_image.shape[1]}")
                if is_multi_channel:
                    if input_image.shape[0] <= 1:
                        raise CellProfilerInputValidationError(f"Expected a 5D array with at least 2 c channels, got {input_image.shape[0]}")
                    
            else: # 2d image. 
                if is_multi_channel: # color
                    if input_image.ndim != 3:
                        raise CellProfilerInputValidationError(f"Expected a 3D array (yxc),got {input_image.ndim}D")
                else: # grayscale
                    if input_image.ndim != 2:
                        raise CellProfilerInputValidationError(f"Expected a 2D array (yx),got {input_image.ndim}D")
        if get_origin(dtype) is Union:
            if input_image.dtype not in get_args(dtype):
                raise CellProfilerInputValidationError(f"Expected an array of type {dtype}, got {input_image.dtype}")
        elif input_image.dtype != dtype:
            raise CellProfilerInputValidationError(f"Expected an array of type {dtype}, got {input_image.dtype}")
        if is_tiled:
            pass
        return input_image
    return validate

def validate_object_labels_dense(input_image: npt.NDArray[np.generic]) -> NDArray[np.generic]:
    if input_image.ndim != 6:
        raise ValueError(f"Expected ObjectLabelsDense as an array of shape (n, c, z, t, y, x), got {input_image.shape}")
    return input_image

def validate_object_label_set(label_set: Sequence[Tuple[NDArray[np.int32], NDArray[np.int32]]]) -> Sequence[Tuple[NDArray[np.int32], NDArray[np.int32]]]:
    # label set is a list of 2 tuples
    for label in label_set:
        if type(label) != tuple or len(label) != 2:
            raise ValueError(f"Expected a list of tuples of length 2, got {label}")
        if type(label[0]) != np.ndarray or type(label[1]) != np.ndarray:
            raise ValueError(f"Expected a list of tuples of ndarrays, got {label}")
        if len(label[0].shape) > 3:
            # see cellprofiler_library.functions.segmentation._validate_labels
            raise ValueError(f"Expected labels of shape (y, x) or (z, y, x), got {label[0].shape}")
    return label_set

Pixel = Annotated[Union[np.float32, np.float64], Field(description="Pixel value")]
ObjectLabel =           Annotated[Union[np.int8, np.int16,np.int32], Field(description="Object label")]

Image2DColor =          Annotated[NDArray[Pixel],       Field(description="2D image with multiple channels of type float32"), AfterValidator(create_type_validator(False, True, False, Union[np.float32, np.float64]))]
Image2DColorMask =      Annotated[NDArray[np.bool_],    Field(description="2D color mask"),        AfterValidator(create_type_validator(False, True, False, np.bool_))]
Image2DGrayscale =      Annotated[NDArray[Pixel],       Field(description="2D grayscale image of type float32"), AfterValidator(create_type_validator(False, False, False, Union[np.float32, np.float64]))]
Image2DGrayscaleMask =  Annotated[NDArray[np.bool_],    Field(description="2D grayscale mask"),    AfterValidator(create_type_validator(False, False, False, np.bool_))]
Image2DBinary =         Annotated[NDArray[np.bool_],    Field(description="2D binary image"),      AfterValidator(create_type_validator(False, False, False, np.bool_))]
Image2DBinaryMask =     Annotated[NDArray[np.bool_],    Field(description="2D binary mask"),       AfterValidator(create_type_validator(False, False, False, np.bool_))]
Image2DInt =            Annotated[NDArray[np.int32],    Field(description="2D int32 image"),       AfterValidator(create_type_validator(False, False, False, np.int32))]

Image3DColor =          Annotated[NDArray[Pixel],       Field(description="3D image with multiple channels of type float32"), AfterValidator(create_type_validator(True, True, False, Union[np.float32, np.float64]))]
Image3DColorMask =      Annotated[NDArray[np.bool_],    Field(description="3D image with multiple channels of type float32"), AfterValidator(create_type_validator(True, True, False, np.bool_))]
Image3DGrayscale =      Annotated[NDArray[Pixel],       Field(description="3D grayscale image of type float32"), AfterValidator(create_type_validator(True, False, False, Union[np.float32, np.float64]))]
Image3DGrayscaleMask =  Annotated[NDArray[np.bool_],    Field(description="3D grayscale mask"),    AfterValidator(create_type_validator(True, False, False, np.bool_))]
Image3DBinary =         Annotated[NDArray[np.bool_],    Field(description="3D binary image"),      AfterValidator(create_type_validator(True, False, False, np.bool_))]
Image3DBinaryMask =     Annotated[NDArray[np.bool_],    Field(description="3D binary mask"),       AfterValidator(create_type_validator(True, False, False, np.bool_))]
Image3DInt =            Annotated[NDArray[np.int32],    Field(description="3D int32 image"),       AfterValidator(create_type_validator(True, False, False, np.int32))]

# see cellprofiler_library.functions.segmentation._validate_<type> for more details 
ObjectLabelsDense =     Annotated[NDArray[ObjectLabel], Field(description="Dense array of object labels"), AfterValidator(validate_object_labels_dense)]
ObjectLabelSet =        Annotated[Sequence[Tuple[NDArray[ObjectLabel], NDArray[np.int32]]], Field(description="List of Tuples of object labels and object numbers in each label matrix"), AfterValidator(validate_object_label_set)]
ObjectSegmentation =    Annotated[NDArray[ObjectLabel], Field(description="Object segmentation")]

ImageGrayscale =        Union[Image2DGrayscale, Image3DGrayscale]
ImageGrayscaleMask =    Union[Image2DGrayscaleMask, Image3DGrayscaleMask]

ImageColor =            Union[Image2DColor, Image3DColor]
ImageColorMask =        Union[Image2DColorMask, Image3DColorMask]

ImageBinary =           Union[Image2DBinary, Image3DBinary]
ImageBinaryMask =       Union[Image2DBinaryMask, Image3DBinaryMask]

ImageAny =              Union[Image2DColor, Image3DColor, Image2DGrayscale, Image3DGrayscale, Image2DBinary, Image3DBinary, Image2DInt, Image3DInt]
ImageAnyMask =          Union[Image2DColorMask, Image3DColorMask, Image2DGrayscaleMask, Image3DGrayscaleMask, Image2DBinaryMask, Image3DBinaryMask]

Image2D =               Union[Image2DColor, Image2DGrayscale, Image2DBinary]
Image2DMask =           Union[Image2DColorMask, Image2DGrayscaleMask]

Image3D =               Union[Image3DColor, Image3DGrayscale, Image3DBinary]
Image3DMask =           Union[Image3DColorMask, Image3DGrayscaleMask]

ImageInt =              Union[Image2DInt, Image3DInt]

StructuringElement =    NDArray[np.uint8]



import warnings

import numpy
import centrosome
from pydantic import Field, validate_call, ConfigDict
from typing import Annotated, Optional
from ..functions.image_processing import (
    enhance_edges_sobel,
    enhance_edges_log,
    enhance_edges_prewitt,
    enhance_edges_canny,
)
from ..opts.enhanceedges import EdgeFindingMethod, EdgeDirection
from ..types import Image2DGrayscale, Image2DGrayscaleMask

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def enhanceedges(
    image:                          Annotated[Image2DGrayscale, Field(description="Input image")],
    mask:                           Annotated[Optional[Image2DGrayscaleMask], Field(description="Mask of the input image")] = None,
    method:                         Annotated[EdgeFindingMethod, Field(description="Edge finding method")] = EdgeFindingMethod.SOBEL,
    automatic_threshold:            Annotated[bool, Field(description="Automatically calculate the threshold?")] = True,
    direction:                      Annotated[EdgeDirection, Field(description="Select edge direction to enhance")] = EdgeDirection.ALL,
    sigma:                          Annotated[int, Field(description="Gaussian's sigma value")] = 10,
    manual_threshold:               Annotated[float, Field(description="Absolute threshold")] = 0.2,
    threshold_adjustment_factor:    Annotated[float, Field(description="Threshold adjustment factor")] = 1.0,
    automatic_low_threshold:        Annotated[bool, Field(description="Automatically calculate the low / soft threshold cutoff for the Canny method?")] = True,
    low_threshold:                  Annotated[float, Field(description="Low threshold value")] = 0.1,
) -> Image2DGrayscale:
    """EnhanceEdges module

    Parameters
    ----------
    image : numpy.array
        Input image
    mask : numpy.array, optional
        Boolean mask, by default None
    method : str, optional
        Enhance edges algorithm to apply to the input image, by default "sobel"
    direction : str, optional
        Applicable to only the Sobel and Prewitt algorithms, by default "all"
    sigma : int, optional
        Applicable to only the Canny and Laplacian of Gaussian algorithms, by default 10. Only considered if automatic_gaussian is False.
    automatic_threshold : bool, optional
        Applicable only to the Canny algorithm, by default True
    manual_threshold : float, optional
        Applicable only to the Canny algorithm, by default 0.2
    threshold_adjustment_factor : float, optional
        Applicable only to the Canny algorithm, by default 1.0
    automatic_low_threshold : bool, optional
        Applicable only to the Canny algorithm, by default True
    low_threshold : float, optional
        Applicable only to the Canny algorithm, by default 0.1

    Returns
    -------
    numpy.array
        Image with enhanced edges
    """

    if not 0 <= low_threshold <= 1:
        warnings.warn(
            f"""low_threshold value of {low_threshold} is outside
            of the [0-1] CellProfiler default."""
        )

    if mask is None:
        mask = numpy.ones(image.shape, bool)

    if method == EdgeFindingMethod.SOBEL:
        output_pixels = enhance_edges_sobel(image, mask, direction)
    elif method == EdgeFindingMethod.LOG:
        output_pixels = enhance_edges_log(image, mask, sigma)
    elif method == EdgeFindingMethod.PREWITT:
        output_pixels = enhance_edges_prewitt(image, mask, direction)
    elif method == EdgeFindingMethod.CANNY:
        output_pixels = enhance_edges_canny(
            image,
            mask,
            auto_threshold=automatic_threshold,
            auto_low_threshold=automatic_low_threshold,
            sigma=sigma,
            low_threshold=low_threshold,
            manual_threshold=manual_threshold,
            threshold_adjustment_factor=threshold_adjustment_factor,
        )
    elif method == EdgeFindingMethod.ROBERTS:
        output_pixels = centrosome.filter.roberts(image, mask)
    elif method == EdgeFindingMethod.KIRSCH:
        output_pixels = centrosome.kirsch.kirsch(image)
    else:
        raise NotImplementedError(f"{method} edge detection method is not implemented.")

    return output_pixels



from enum import Enum

class SecondaryObjectMethod(str, Enum):
    PROPAGATION = "Propagation"
    WATERSHED_GRADIENT = "Watershed - Gradient"
    WATERSHED_IMAGE = "Watershed - Image"
    DISTANCE_N = "Distance - N"
    DISTANCE_B = "Distance - B"


from typing import Any, Dict, List
from pydantic import BaseModel, Field, ConfigDict

class LibraryMeasurements(BaseModel):
    """
    A Pydantic implementation mirroring CellProfiler's Measurements class functionality.
    
    This class represents measurements for a single image set context, designed to be
    constructed from and serializable to the "Primitive Dictionary" contract.
    """
    model_config = ConfigDict(
        arbitrary_types_allowed=True, 
        populate_by_name=True
    )

    # Primary data structures matching the MDC contract
    image: Dict[str, Any] = Field(default_factory=dict)
    objects: Dict[str, Dict[str, Any]] = Field(default_factory=dict)
    
    # Additional storage for full Measurements compatibility (e.g. Experiment metadata)
    experiment: Dict[str, Any] = Field(default_factory=dict)
    
    # Placeholder for relationships if needed in the future
    relationships: List[Dict[str, Any]] = Field(default_factory=list)

    def add_measurement(self, object_name: str, feature_name: str, data: Any):
        """
        Add a measurement. Mirrors Measurements.add_measurement.
        
        Args:
            object_name: Name of the object (e.g., 'Image', 'Nuclei').
            feature_name: Name of the feature (e.g., 'AreaShape_Area').
            data: The value to store (float, string, or numpy array).
        """
        if object_name == "Experiment":
            self.experiment[feature_name] = data
        elif object_name == "Image":
            self.image[feature_name] = data
        else:
            if object_name not in self.objects:
                self.objects[object_name] = {}
            self.objects[object_name][feature_name] = data

    def get_measurement(self, object_name: str, feature_name: str) -> Any:
        """
        Get a measurement. Mirrors Measurements.get_measurement.
        """
        if object_name == "Experiment":
            return self.experiment.get(feature_name)
        elif object_name == "Image":
            return self.image.get(feature_name)
        else:
            return self.objects.get(object_name, {}).get(feature_name)

    def add_image_measurement(self, feature_name: str, data: Any):
        """Helper to add an image measurement."""
        self.add_measurement("Image", feature_name, data)

    def add_experiment_measurement(self, feature_name: str, data: Any):
        """Helper to add an experiment measurement."""
        self.add_measurement("Experiment", feature_name, data)

    def get_experiment_measurement(self, feature_name: str) -> Any:
        """Helper to get an experiment measurement."""
        return self.get_measurement("Experiment", feature_name)
    
    def get_object_names(self) -> List[str]:
        """Return a list of all object names (including Image and Experiment if present)."""
        names = list(self.objects.keys())
        if self.image:
            names = ["Image"] + names
        if self.experiment:
            names = ["Experiment"] + names
        return names

    def get_feature_names(self, object_name: str) -> List[str]:
        """Return feature names for a specific object."""
        if object_name == "Experiment":
            return list(self.experiment.keys())
        elif object_name == "Image":
            return list(self.image.keys())
        else:
            return list(self.objects.get(object_name, {}).keys())
            
    def has_feature(self, object_name: str, feature_name: str) -> bool:
        """Check if a feature exists."""
        if object_name == "Experiment":
            return feature_name in self.experiment
        elif object_name == "Image":
            return feature_name in self.image
        else:
            return feature_name in self.objects.get(object_name, {})

    def __getitem__(self, key):
        """Support dict-style access: m['Image', 'Count_Cells']"""
        if isinstance(key, tuple):
             assert len(key) == 2
             return self.get_measurement(*key)
        raise KeyError("Invalid key format. Expected (Object, Feature)")
        
    def __setitem__(self, key, value):
        """Support dict-style assignment: m['Image', 'Count_Cells'] = 5"""
        if isinstance(key, tuple):
             assert len(key) == 2
             self.add_measurement(*key, value)
        else:
             raise KeyError("Invalid key format. Expected (Object, Feature)")

    def to_dict(self) -> Dict[str, Any]:
        """
        Serialize to the primitive dictionary format required by the refactoring contract.
        """
        return self.model_dump(exclude_none=True)

    def merge(self, other: 'LibraryMeasurements') -> 'LibraryMeasurements':
        """
        Merge this LibraryMeasurements object with another one, returning a new instance.
        
        Collision resolution:
        - Image measurements: 'other' overwrites 'self'
        - Experiment measurements: 'other' overwrites 'self'
        - Object measurements: Merged by object name. Within an object, 'other' features overwrite 'self'.
        - Relationships: Concatenated
        
        Args:
            other: Another LibraryMeasurements instance
            
        Returns:
            A new LibraryMeasurements instance containing merged data
        """
        # Create deep copies of dictionaries to avoid side effects
        new_image = self.image.copy()
        new_image.update(other.image)
        
        new_experiment = self.experiment.copy()
        new_experiment.update(other.experiment)
        
        new_objects = {}
        # Merge objects
        all_objects = set(self.objects.keys()) | set(other.objects.keys())
        for obj_name in all_objects:
            obj_measurements = self.objects.get(obj_name, {}).copy()
            obj_measurements.update(other.objects.get(obj_name, {}))
            new_objects[obj_name] = obj_measurements
            
        # Concatenate relationships
        new_relationships = self.relationships + other.relationships
        
        return LibraryMeasurements(
            image=new_image,
            objects=new_objects,
            experiment=new_experiment,
            relationships=new_relationships
        )



import numpy
from typing import Annotated, Optional, Tuple, Callable, Dict, Union
from pydantic import Field, validate_call, ConfigDict
from cellprofiler_library.opts.convertobjectstoimage import ImageMode
from cellprofiler_library.types import ImageBinary, ImageColor, ImageGrayscale, ObjectLabelSet, ImageAny, ImageInt
from cellprofiler_library.functions.object_processing import image_mode_black_and_white, image_mode_grayscale, image_mode_color, image_mode_uint16

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def convert_objects_to_image(
        image_mode: Annotated[ImageMode, Field(description="Color format to be used for conversion")],
        objects_labels : Annotated[ObjectLabelSet, Field(description="Labels of the objects")],
        objects_shape : Annotated[Tuple[int, ...], Field(description="Shape of the objects")],
        colormap_value : Annotated[Optional[str], Field(description="Colormap to be used for conversion")] = None
        ) -> ImageAny:

    alpha = numpy.zeros(objects_shape, numpy.int32)

    converter_fn_map = {
        ImageMode.BINARY: image_mode_black_and_white,
        ImageMode.GRAYSCALE: image_mode_grayscale,
        ImageMode.COLOR: image_mode_color,
        ImageMode.UINT16: image_mode_uint16,
    }

    pixel_data_init_map: Dict[
        ImageMode,
        Callable[[], Union[ImageGrayscale, ImageBinary, ImageColor, ImageInt]]
    ] = {
        ImageMode.BINARY: lambda: numpy.zeros(objects_shape, bool),
        ImageMode.GRAYSCALE: lambda: numpy.zeros(objects_shape),
        ImageMode.COLOR: lambda: numpy.zeros(objects_shape + (3,)),
        ImageMode.UINT16: lambda: numpy.zeros(objects_shape, numpy.int32),
    }
    pixel_data = pixel_data_init_map.get(image_mode, lambda: numpy.zeros(objects_shape + (3,)))()
    for labels, _ in objects_labels:
        mask = labels != 0
        if numpy.all(~mask):
            continue
        pixel_data, alpha = converter_fn_map[image_mode](pixel_data, mask, alpha, labels, colormap_value)
    mask = alpha > 0
    if image_mode == ImageMode.COLOR:
        pixel_data[mask, :] = pixel_data[mask, :] / alpha[mask][:, numpy.newaxis]
    elif image_mode != ImageMode.BINARY:
        pixel_data[mask] = pixel_data[mask] / alpha[mask]
    return pixel_data



from ..functions.image_processing import overlay_objects

def overlayobjects(
    image, labels, opacity=0.3, max_label=None, seed=None, colormap="jet"
):
    return overlay_objects(
        image=image, 
        labels=labels, 
        opacity=opacity, 
        max_label=max_label, 
        seed=seed,
        colormap=colormap
        )


from enum import Enum

class OperationMethod(str, Enum):
    ENHANCE = "Enhance"
    SUPPRESS = "Suppress"

class EnhanceMethod(str, Enum):
    SPECKLES = "Speckles"
    NEURITES = "Neurites"
    DARK_HOLES = "Dark holes"
    CIRCLES = "Circles"
    TEXTURE = "Texture"
    DIC = "DIC"

class SpeckleAccuracy(str, Enum):
    SLOW = "Slow"
    FAST = "Fast"

class NeuriteMethod(str, Enum):
    GRADIENT = "Line structures"
    TUBENESS = "Tubeness"



from enum import Enum

class UnclumpMethod(str, Enum):
    INTENSITY = "Intensity"
    SHAPE = "Shape"
    NONE = "None"

class WatershedMethod(str, Enum):
    INTENSITY = "Intensity"
    SHAPE = "Shape"
    PROPAGATE = "Propagate"
    NONE = "None"

class FillHolesMethod(str, Enum):
    NEVER = "Never"
    THRESHOLDING = "After both thresholding and declumping"
    DECLUMP = "After declumping only"
FH_ALL = (FillHolesMethod.NEVER.value, FillHolesMethod.THRESHOLDING.value, FillHolesMethod.DECLUMP.value)

class LimitObjects(str, Enum):
    NONE = "Continue"
    ERASE = "Erase"



from enum import Enum


class MorphFunction(str, Enum):
    """Morphological operations available in the Morph module."""
    BRANCHPOINTS = "branchpoints"
    BRIDGE = "bridge"
    CLEAN = "clean"
    CONVEX_HULL = "convex hull"
    DIAG = "diag"
    DISTANCE = "distance"
    ENDPOINTS = "endpoints"
    FILL = "fill"
    HBREAK = "hbreak"
    MAJORITY = "majority"
    OPENLINES = "openlines"
    REMOVE = "remove"
    SHRINK = "shrink"
    SKELPE = "skelpe"
    SPUR = "spur"
    THICKEN = "thicken"
    THIN = "thin"
    VBREAK = "vbreak"

MorphFunctions_All = [
    MorphFunction.BRANCHPOINTS.value,
    MorphFunction.BRIDGE.value,
    MorphFunction.CLEAN.value,
    MorphFunction.CONVEX_HULL.value,
    MorphFunction.DIAG.value,
    MorphFunction.DISTANCE.value,
    MorphFunction.ENDPOINTS.value,
    MorphFunction.FILL.value,
    MorphFunction.HBREAK.value,
    MorphFunction.MAJORITY.value,
    MorphFunction.OPENLINES.value,
    MorphFunction.REMOVE.value,
    MorphFunction.SHRINK.value,
    MorphFunction.SKELPE.value,
    MorphFunction.SPUR.value,
    MorphFunction.THICKEN.value,
    MorphFunction.THIN.value,
    MorphFunction.VBREAK.value,
]



import numpy
from typing import Annotated, Optional
from pydantic import Field, validate_call, ConfigDict

from cellprofiler_library.functions.image_processing import smoothing_gaussian, smoothing_median, smoothing_keeping_edges, smoothing_fit_polynomial, smoothing_circular_average, smoothing_smooth_to_average
from cellprofiler_library.opts.smooth import SmoothingMethod
from cellprofiler_library.types import Image2D, Image2DMask


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def smooth(
        pixel_data:         Annotated[Image2D, Field(description="Pixel data of image to be smoothed")],
        mask:               Annotated[Optional[Image2DMask], Field(description="Mask of the pixel data")],
        multichannel:       Annotated[bool, Field(description="Set to true if image is multichannel")],
        object_size:        Annotated[Optional[float], Field(description="Object size")],
        smoothing_method:   Annotated[SmoothingMethod, Field(description="Smoothing method")],
        sigma_range:        Annotated[Optional[float], Field(description="Sigma range")],
        clip:               Annotated[Optional[bool], Field(description="Clip intensities to 0 and 1")] = True,
) -> Image2D:
    if object_size is not None:
        obj_size = float(object_size)
    else:
        obj_size = float(min(30, max(1, numpy.mean(pixel_data.shape) / 40)))
    sigma = obj_size / 2.35

    if smoothing_method == SmoothingMethod.GAUSSIAN_FILTER:
        output_pixels = smoothing_gaussian(pixel_data, mask, sigma)
    elif smoothing_method == SmoothingMethod.MEDIAN_FILTER:
        output_pixels = smoothing_median(pixel_data, mask, obj_size)
    elif smoothing_method == SmoothingMethod.SMOOTH_KEEPING_EDGES:
        assert sigma_range is not None, "sigma_range must be provided for smooth_keeping_edges"
        output_pixels = smoothing_keeping_edges(pixel_data, multichannel, sigma_range, sigma)
    elif smoothing_method == SmoothingMethod.FIT_POLYNOMIAL:
        assert clip is not None, "clip must be provided for fit_polynomial"
        output_pixels = smoothing_fit_polynomial(pixel_data, mask, clip)
    elif smoothing_method == SmoothingMethod.CIRCULAR_AVERAGE_FILTER:
        output_pixels = smoothing_circular_average(pixel_data, mask, obj_size)
    elif smoothing_method == SmoothingMethod.SM_TO_AVERAGE:
        output_pixels = smoothing_smooth_to_average(pixel_data, mask)
    else:
        raise ValueError(
            "Unsupported smoothing method: %s" % smoothing_method
        )
    return output_pixels


from ..functions.image_processing import morphology_closing


def closing(image, structuring_element):
    return morphology_closing(
        image,
        structuring_element=structuring_element,
    )



# coding=utf-8

"""
DilateObjects module for the CellProfiler library.

This module contains the core algorithms for object dilation operations.
"""

from pydantic import validate_call, ConfigDict, Field
from typing import Union, Tuple, Annotated
from cellprofiler_library.types import StructuringElement, ObjectSegmentation
from cellprofiler_library.functions.object_processing import dilate_objects_with_structuring_element
from cellprofiler_library.functions.image_processing import get_structuring_element
from cellprofiler_library.opts.structuring_elements import StructuringElementShape2D, StructuringElementShape3D

StructuringElementSize = Annotated[int, Field(description="Size of structuring element", gt=0)]
StructuringElementParameters = Tuple[Union[StructuringElementShape2D, StructuringElementShape3D], StructuringElementSize]

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def dilate_objects(
    labels: Annotated[ObjectSegmentation, Field(description="Input object segmentations")],
    structuring_element: Annotated[Union[StructuringElement, StructuringElementParameters], Field(description="Structuring element for dilation operation as either an NDArray or a tuple of (StructuringElement[N]D, size)")],
) -> ObjectSegmentation:
    """Dilate objects based on the structuring element provided.
    
    This function is similar to the "Expand" function of ExpandOrShrinkObjects,
    with two major distinctions:
    1. DilateObjects supports 3D objects, unlike ExpandOrShrinkObjects.
    2. In ExpandOrShrinkObjects, two objects closer than the expansion distance
       will expand until they meet and then stop there. In this module, the object with
       the larger object number (the object that is lower in the image) will be expanded
       on top of the object with the smaller object number.
    
    Args:
        labels: Input labeled objects array
        structuring_element: Structuring element for dilation operation
        
    Returns:
        Dilated objects array with same dimensions as input
    """
    if isinstance(structuring_element, tuple):
        structuring_element = get_structuring_element(structuring_element[0], structuring_element[1])
    return dilate_objects_with_structuring_element(
        labels=labels,
        structuring_element=structuring_element
    )



from enum import Enum

class FlipDirection(str, Enum):
    NONE = "Do not flip"
    LEFT_TO_RIGHT = "Left to right"
    TOP_TO_BOTTOM = "Top to bottom"
    BOTH = "Left to right and top to bottom"

class RotateMethod(str, Enum):
    NONE = "Do not rotate"
    ANGLE = "Enter angle"
    COORDINATES = "Enter coordinates"

class RotationCoordinateAlignmnet(str, Enum):
    HORIZONTALLY = "Horizontally"
    VERTICALLY = "Vertically"

D_ANGLE = "angle"

"""Rotation measurement category"""
M_ROTATION_CATEGORY = "Rotation"
"""Rotation measurement format (+ image name)"""
M_ROTATION_F = "%s_%%s" % M_ROTATION_CATEGORY

FLIP_ALL = [FlipDirection.NONE, FlipDirection.LEFT_TO_RIGHT, FlipDirection.TOP_TO_BOTTOM, FlipDirection.BOTH]

ROTATE_ALL = [RotateMethod.NONE, RotateMethod.ANGLE, RotateMethod.COORDINATES]

C_ALL = [RotationCoordinateAlignmnet.HORIZONTALLY, RotationCoordinateAlignmnet.VERTICALLY]



from pydantic import validate_call, ConfigDict, Field
from typing import Annotated
from cellprofiler_library.types import ObjectSegmentation
from cellprofiler_library.functions.object_processing import find_centroids

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def shrink_to_object_centers(
    label_image: Annotated[ObjectSegmentation, Field(description="Input label image")]
    ) -> ObjectSegmentation:
    return find_centroids(label_image)


from enum import Enum

class BrightnessMode(str, Enum):
    MAX_IMAGE = "Max of image"
    MAX_POSSIBLE = "Max possible"


class LineMode(str, Enum):
    INNER = "Inner"
    OUTER = "Outer"
    THICK = "Thick"


class SourceMode(str, Enum):
    FROM_IMAGES = "Image"
    FROM_OBJECTS = "Objects"

class OutlineMode(str, Enum):
    COLOR = "Color"
    GRAYSCALE = "Grayscale"


# Color definitions - preserved exactly from original module
COLORS = {
    "White": (1, 1, 1),
    "Black": (0, 0, 0),
    "Red": (1, 0, 0),
    "Green": (0, 1, 0),
    "Blue": (0, 0, 1),
    "Yellow": (1, 1, 0),
}

# Color order - preserved exactly from original module
COLOR_ORDER = ["Red", "Green", "Blue", "Yellow", "White", "Black"]



from enum import Enum

class Shape(str, Enum):
    RECTANGLE = "Rectangle"
    ELLIPSE = "Ellipse"
    IMAGE = "Image"
    OBJECTS = "Objects"
    CROPPING = "Previous cropping"

class RemovalMethod(str, Enum):
    NO = "No"
    EDGES = "Edges"
    ALL = "All"

class Measurement(str, Enum):
    AREA_RETAINED = "Crop_AreaRetainedAfterCropping_%s"
    ORIGINAL_AREA = "Crop_OriginalImageArea_%s"

class CroppingMethod(str, Enum):
    COORDINATES = "Coordinates"
    MOUSE = "Mouse"

class CroppingPattern(str, Enum):
    FIRST = "First"
    INDIVIDUALLY = "Individually"

class Limits(str, Enum):
    ABSOLUTE = "Absolute"
    FROM_EDGE = "From edge"

class Ellipse(str, Enum):
    XCENTER = "xcenter"
    YCENTER = "ycenter"
    XRADIUS = "xradius"
    YRADIUS = "yradius"

class Rectangle(str, Enum):
    LEFT = "left"
    TOP = "top"
    RIGHT = "right"
    BOTTOM = "bottom"



from ..functions.image_processing import reduce_noise

def reducenoise(image, patch_size, patch_distance, cutoff_distance, channel_axis=None):
    denoised = reduce_noise(
        image,
        patch_size=patch_size,
        patch_distance=patch_distance,
        cutoff_distance=cutoff_distance,
        channel_axis=channel_axis,
    )
    return denoised



# coding=utf-8

"""
ErodeObjects module for the CellProfiler library.

This module contains the core algorithms for object erosion operations.
"""

from pydantic import validate_call, ConfigDict, Field
from typing import Union, Tuple, Annotated
from cellprofiler_library.types import StructuringElement, ObjectSegmentation
from cellprofiler_library.functions.object_processing import erode_objects_with_structuring_element
from cellprofiler_library.functions.image_processing import get_structuring_element
from cellprofiler_library.opts.structuring_elements import StructuringElementShape2D, StructuringElementShape3D

StructuringElementSize = Annotated[int, Field(description="Size of structuring element", gt=0)]
StructuringElementParameters = Tuple[Union[StructuringElementShape2D, StructuringElementShape3D], StructuringElementSize]

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def erode_objects(
    labels: Annotated[ObjectSegmentation, Field(description="Input object segmentations")],
    structuring_element: Annotated[Union[StructuringElement, StructuringElementParameters], Field(description="Structuring element for erosion operation as either an NDArray or a tuple of (StructuringElement[N]D, size)")],
    preserve_midpoints: Annotated[bool, Field(description="If set to True, the central pixels for each object will not be eroded. This ensures that objects are not lost.")] = False,
    relabel_objects: Annotated[bool, Field(description="Selecting True will assign new label numbers to resulting objects")] = False
) -> ObjectSegmentation:
    """Erode objects based on the structuring element provided.
    
    This function is similar to the "Shrink" function of ExpandOrShrinkObjects,
    with two major distinctions:
    1. ErodeObjects supports 3D objects, unlike ExpandOrShrinkObjects.
    2. An object smaller than the structuring element will be removed entirely
       unless preserve_midpoints is enabled.
    
    Args:
        labels: Input labeled objects array
        structuring_element: Structuring element for erosion operation
        preserve_midpoints: If set to True, the central pixels for each object will not be eroded. This ensures that objects are not lost.
        relabel_objects: If set to True, the resulting objects will be relabeled with new label numbers
        
    Returns:
        Eroded objects array with same dimensions as input
    """
    if isinstance(structuring_element, tuple):
        structuring_element = get_structuring_element(structuring_element[0], structuring_element[1])
    return erode_objects_with_structuring_element(
        labels=labels,
        structuring_element=structuring_element,
        preserve_midpoints=preserve_midpoints,
        relabel_objects=relabel_objects
    )



import numpy
from typing import Annotated, Union, Tuple
from pydantic import validate_call, ConfigDict, Field
from cellprofiler_library.functions.object_processing import outline
from cellprofiler_library.types import ObjectSegmentation

@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def identifytertiaryobjects(
        primary_labels:     Annotated[ObjectSegmentation, Field(description="Primary object segmentations")] = None,
        secondary_labels:   Annotated[ObjectSegmentation, Field(description="Secondary object segmentations")] = None,
        shrink_primary:     Annotated[bool, Field(description="Shrink the primary objects")] = True,
        return_cp_output:   Annotated[bool, Field(description="Return CellProfiler output")] = False
) -> Union[
    ObjectSegmentation,
    Tuple[
        ObjectSegmentation,
        ObjectSegmentation
    ]
]:
    # If size/shape differences were too extreme, raise an error.
    if primary_labels.shape != secondary_labels.shape:
        raise ValueError(
            f"""
            This module requires that the object sets have matching widths
            and matching heights. The primary and secondary objects do not
            ({primary_labels.shape} vs {secondary_labels.shape}). If they
            are paired correctly you may want to use the ResizeObjects
            module to make them the same size.
            """
        )
    #
    # Find the outlines of the primary image and use this to shrink the
    # primary image by one. This guarantees that there is something left
    # of the secondary image after subtraction
    #
    primary_outline = outline(primary_labels)
    tertiary_labels = secondary_labels.copy()
    if shrink_primary:
        primary_mask = numpy.logical_or(primary_labels == 0, primary_outline)
    else:
        primary_mask = primary_labels == 0
    tertiary_labels[primary_mask == False] = 0
     #
    # Check if a label was deleted as a result of the subtraction
    #
    secondary_unique_labels, secondary_unique_indices = numpy.unique(secondary_labels, return_index=True)
    tertiary_unique_labels = numpy.unique(tertiary_labels)
    missing_labels = numpy.setdiff1d(secondary_unique_labels, tertiary_unique_labels)
    for missing_label in missing_labels:
        # If a label was deleted, manually add a pixel to the tertiary_labels.
        # This workaround ensures that ghost objects do not get created by identifytertiaryobjects.
        
        # first non-zero (top-left) coodrinate of the secondary object is used to add a pixel to the tertiary_labels
        first_row, first_col = numpy.unravel_index(secondary_unique_indices[missing_label], secondary_labels.shape)
        tertiary_labels[first_row, first_col] = missing_label
    
    if return_cp_output:
        #
        # Get the outlines of the tertiary image
        #
        tertiary_outlines = outline(tertiary_labels) != 0
        return (
            tertiary_labels,
            tertiary_outlines
        )
    return tertiary_labels


import logging
from typing import Annotated, Optional, List, Dict, Any
from pydantic import validate_call, Field, ConfigDict

from ..opts.morph import MorphFunction
from ..functions.image_processing import (
    apply_branchpoints,
    apply_bridge, 
    apply_clean,
    apply_convex_hull,
    apply_diag,
    apply_distance,
    apply_endpoints,
    apply_fill,
    apply_hbreak,
    apply_majority,
    apply_openlines,
    apply_remove,
    apply_shrink,
    apply_skelpe,
    apply_spur,
    apply_thicken,
    apply_thin,
    apply_vbreak,
    MorphImageT
)
from ..types import ImageGrayscaleMask

LOGGER = logging.getLogger(__name__)

def morph_operation(
        pixel_data:     Annotated[MorphImageT, Field(description="Image array supporting binary, integer, or float types")],
        mask:           Annotated[Optional[ImageGrayscaleMask], Field(description="Optional boolean mask array")],
        function_name:  Annotated[MorphFunction, Field(description="Morphological operation to perform")],
        repeat_count:   Annotated[int, Field(description="Number of times to repeat the operation", ge=1)],
        custom_repeats: Annotated[int, Field(description="Custom repeat value for specific operations", ge=1)],
        rescale_values: Annotated[bool, Field(description="Whether to rescale distance values from 0 to 1")],
        ) -> MorphImageT:
    """Apply a morphological operation to the image, routing to appropriate function from functions layer."""
    count = repeat_count

    is_binary = pixel_data.dtype.kind == "b"

    if (
        function_name != MorphFunction.OPENLINES and not is_binary
    ):
        # Apply a very crude threshold to the image for binary algorithms
        LOGGER.warning(
            "Warning: converting image to binary for %s\n" % function_name
        )
        pixel_data = pixel_data != 0

    # Route to appropriate function from functions layer
    if function_name == MorphFunction.BRANCHPOINTS:
        return apply_branchpoints(pixel_data, mask)
    elif function_name == MorphFunction.BRIDGE:
        return apply_bridge(pixel_data, mask, count)
    elif function_name == MorphFunction.CLEAN:
        return apply_clean(pixel_data, mask, count)
    elif function_name == MorphFunction.CONVEX_HULL:
        return apply_convex_hull(pixel_data, mask)
    elif function_name == MorphFunction.DIAG:
        return apply_diag(pixel_data, mask, count)
    elif function_name == MorphFunction.DISTANCE:
        return apply_distance(pixel_data, rescale_values)
    elif function_name == MorphFunction.ENDPOINTS:
        return apply_endpoints(pixel_data, mask)
    elif function_name == MorphFunction.FILL:
        return apply_fill(pixel_data, mask, count)
    elif function_name == MorphFunction.HBREAK:
        return apply_hbreak(pixel_data, mask, count)
    elif function_name == MorphFunction.MAJORITY:
        return apply_majority(pixel_data, mask, count)
    elif function_name == MorphFunction.OPENLINES:
        return apply_openlines(pixel_data, mask, custom_repeats)
    elif function_name == MorphFunction.REMOVE:
        return apply_remove(pixel_data, mask, count)
    elif function_name == MorphFunction.SHRINK:
        return apply_shrink(pixel_data, count)
    elif function_name == MorphFunction.SKELPE:
        return apply_skelpe(pixel_data, mask)
    elif function_name == MorphFunction.SPUR:
        return apply_spur(pixel_data, mask, count)
    elif function_name == MorphFunction.THICKEN:
        return apply_thicken(pixel_data, mask, count)
    elif function_name == MorphFunction.THIN:
        return apply_thin(pixel_data, mask, count)
    elif function_name == MorphFunction.VBREAK:
        return apply_vbreak(pixel_data, mask)
    else:
        raise NotImplementedError(
            "Unimplemented morphological function: %s" % function_name
        )


@validate_call(config=ConfigDict(arbitrary_types_allowed=True))
def apply_morphological_operations(
        pixel_data:         Annotated[MorphImageT, Field(description="Image array supporting binary, integer, or float types")],
        mask:               Annotated[Optional[ImageGrayscaleMask], Field(description="Optional boolean mask array")] = None,
        operations_list:    Annotated[List[Dict[str, Any]], Field(description="List of morphological operations to apply sequentially")] = [],
        ) -> MorphImageT:
    """Apply a sequence of morphological operations to the image.
    
    Args:
        pixel_data: Input image pixel data to process
        mask: Optional image mask to constrain operations
        operations_list: List of operation dictionaries, each containing:
            - function_name: Name of morphological operation (MorphFunction enum)
            - repeat_count: Number of times to repeat the operation
            - custom_repeats: Custom repeat value for specific operations
            - rescale_values: Whether to rescale distance values from 0 to 1
    
    Returns:
        Processed pixel data array with applied morphological operations
    """
    result_pixel_data = pixel_data
    
    for operation in operations_list:
        result_pixel_data = morph_operation(
            result_pixel_data,
            mask,
            operation["function_name"],
            operation["repeat_count"],
            operation["custom_repeats"],
            operation["rescale_values"]
        )
    
    return result_pixel_data
